# systemctl-mqtt - MQTT client triggering & reporting shutdown on systemd-based systems
#
# Copyright (C) 2024 Fabian Peter Hammerle <fabian@hammerle.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import jeepney
import systemctl_mqtt._dbus

_LOGGER = logging.getLogger(__name__)


class ServiceManager(jeepney.MessageGenerator):
    """
    https://www.freedesktop.org/software/systemd/man/latest/org.freedesktop.systemd1.html
    """

    # pylint: disable=too-few-public-methods

    interface = "org.freedesktop.systemd1.Manager"

    def __init__(self):
        super().__init__(
            object_path="/org/freedesktop/systemd1", bus_name="org.freedesktop.systemd1"
        )

    # pylint: disable=invalid-name

    def GetUnit(self, name: str) -> jeepney.low_level.Message:
        return jeepney.new_method_call(
            remote_obj=self, method="GetUnit", signature="s", body=(name,)
        )

    def StartUnit(self, name: str, mode: str) -> jeepney.low_level.Message:
        return jeepney.new_method_call(
            remote_obj=self,
            method="StartUnit",
            signature="ss",
            body=(
                name,
                mode,
            ),
        )

    def StopUnit(self, name: str, mode: str) -> jeepney.low_level.Message:
        return jeepney.new_method_call(
            remote_obj=self,
            method="StopUnit",
            signature="ss",
            body=(
                name,
                mode,
            ),
        )

    def RestartUnit(self, name: str, mode: str) -> jeepney.low_level.Message:
        return jeepney.new_method_call(
            remote_obj=self,
            method="RestartUnit",
            signature="ss",
            body=(
                name,
                mode,
            ),
        )


class Unit(systemctl_mqtt._dbus.Properties):  # pylint: disable=protected-access
    """
    https://www.freedesktop.org/software/systemd/man/latest/org.freedesktop.systemd1.html#Unit%20Objects
    """

    # pylint: disable=too-few-public-methods

    interface = "org.freedesktop.systemd1.Unit"

    def __init__(self, *, object_path: str):
        super().__init__(object_path=object_path, bus_name="org.freedesktop.systemd1")

    # pylint: disable=invalid-name


def start_unit(unit_name: str):
    proxy = get_service_manager_proxy()
    try:
        proxy.StartUnit(unit_name, "replace")
        _LOGGER.debug("Starting unit: %s", unit_name)
    # pylint: disable=broad-exception-caught
    except jeepney.wrappers.DBusErrorResponse as exc:
        _LOGGER.error("Failed to start unit: %s because %s ", unit_name, exc.name)


def stop_unit(unit_name: str):
    proxy = get_service_manager_proxy()
    try:
        proxy.StopUnit(unit_name, "replace")
        _LOGGER.debug("Stopping unit: %s", unit_name)
    # pylint: disable=broad-exception-caught
    except jeepney.wrappers.DBusErrorResponse as exc:
        _LOGGER.error("Failed to stop unit: %s because %s ", unit_name, exc.name)


def restart_unit(unit_name: str):
    proxy = get_service_manager_proxy()
    try:
        proxy.RestartUnit(unit_name, "replace")
        _LOGGER.debug("Restarting unit: %s", unit_name)
    # pylint: disable=broad-exception-caught
    except jeepney.wrappers.DBusErrorResponse as exc:
        _LOGGER.error("Failed to restart unit: %s because %s ", unit_name, exc.name)


def get_service_manager_proxy() -> jeepney.io.blocking.Proxy:
    # https://jeepney.readthedocs.io/en/latest/integrate.html
    # https://gitlab.com/takluyver/jeepney/-/blob/master/examples/aio_notify.py
    return jeepney.io.blocking.Proxy(
        msggen=ServiceManager(),
        connection=jeepney.io.blocking.open_dbus_connection(
            bus="SYSTEM",
        ),
    )



# systemctl-mqtt - MQTT client triggering & reporting shutdown on systemd-based systems
#
# Copyright (C) 2020 Fabian Peter Hammerle <fabian@hammerle.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime
import json
import logging
import re
import typing
import unittest.mock

import jeepney.wrappers
import pytest

import systemctl_mqtt

# pylint: disable=protected-access


def test_shutdown_lock():
    lock_fd = unittest.mock.MagicMock(spec=jeepney.fds.FileDescriptor)
    with unittest.mock.patch(
        "systemctl_mqtt._dbus.login_manager.get_login_manager_proxy"
    ) as get_login_manager_mock:
        state = systemctl_mqtt._State(
            mqtt_topic_prefix="any",
            homeassistant_discovery_prefix=None,
            homeassistant_discovery_object_id=None,
            poweroff_delay=datetime.timedelta(),
            monitored_system_unit_names=[],
            controlled_system_unit_names=[],
        )
        get_login_manager_mock.return_value.Inhibit.return_value = (lock_fd,)
        state.acquire_shutdown_lock()
    state._login_manager.Inhibit.assert_called_once_with(
        what="shutdown",
        who="systemctl-mqtt",
        why="Report shutdown via MQTT",
        mode="delay",
    )
    assert state._shutdown_lock == lock_fd
    lock_fd.close.assert_not_called()
    state.release_shutdown_lock()
    lock_fd.close.assert_called_once_with()


@pytest.mark.asyncio
@pytest.mark.parametrize("active", [True, False])
async def test_preparing_for_shutdown_handler(active: bool) -> None:
    with unittest.mock.patch(
        "systemctl_mqtt._dbus.login_manager.get_login_manager_proxy"
    ):
        state = systemctl_mqtt._State(
            mqtt_topic_prefix="any",
            homeassistant_discovery_prefix="pre/fix",
            homeassistant_discovery_object_id="obj",
            poweroff_delay=datetime.timedelta(),
            monitored_system_unit_names=[],
            controlled_system_unit_names=[],
        )
    mqtt_client_mock = unittest.mock.MagicMock()
    with unittest.mock.patch.object(
        state, "_publish_preparing_for_shutdown"
    ) as publish_mock, unittest.mock.patch.object(
        state, "acquire_shutdown_lock"
    ) as acquire_lock_mock, unittest.mock.patch.object(
        state, "release_shutdown_lock"
    ) as release_lock_mock:
        await state.preparing_for_shutdown_handler(
            active=active, mqtt_client=mqtt_client_mock
        )
    publish_mock.assert_awaited_once_with(mqtt_client=mqtt_client_mock, active=active)
    if active:
        acquire_lock_mock.assert_not_called()
        release_lock_mock.assert_called_once_with()
    else:
        acquire_lock_mock.assert_called_once_with()
        release_lock_mock.assert_not_called()


@pytest.mark.asyncio
@pytest.mark.parametrize("active", [True, False])
async def test_publish_preparing_for_shutdown(active: bool) -> None:
    login_manager_mock = unittest.mock.MagicMock()
    login_manager_mock.Get.return_value = (("b", active),)[:]
    with unittest.mock.patch(
        "systemctl_mqtt._dbus.login_manager.get_login_manager_proxy",
        return_value=login_manager_mock,
    ):
        state = systemctl_mqtt._State(
            mqtt_topic_prefix="any",
            homeassistant_discovery_prefix="pre/fix",
            homeassistant_discovery_object_id="obj",
            poweroff_delay=datetime.timedelta(),
            monitored_system_unit_names=[],
            controlled_system_unit_names=[],
        )
    assert state._login_manager == login_manager_mock
    mqtt_client_mock = unittest.mock.AsyncMock()
    await state.publish_preparing_for_shutdown(mqtt_client=mqtt_client_mock)
    login_manager_mock.Get.assert_called_once_with("PreparingForShutdown")
    mqtt_client_mock.publish.assert_awaited_once_with(
        topic="any/preparing-for-shutdown",
        payload="true" if active else "false",
        retain=False,
    )


class DBusErrorResponseMock(jeepney.wrappers.DBusErrorResponse):
    # pylint: disable=missing-class-docstring,super-init-not-called
    def __init__(self, name: str, data: typing.Any):
        self.name = name
        self.data = data


@pytest.mark.asyncio
async def test_publish_preparing_for_shutdown_get_fail(caplog):
    login_manager_mock = unittest.mock.MagicMock()
    login_manager_mock.Get.side_effect = DBusErrorResponseMock("error", ("mocked",))
    with unittest.mock.patch(
        "systemctl_mqtt._dbus.login_manager.get_login_manager_proxy",
        return_value=login_manager_mock,
    ):
        state = systemctl_mqtt._State(
            mqtt_topic_prefix="any",
            homeassistant_discovery_prefix=None,
            homeassistant_discovery_object_id=None,
            poweroff_delay=datetime.timedelta(),
            monitored_system_unit_names=[],
            controlled_system_unit_names=[],
        )
    mqtt_client_mock = unittest.mock.MagicMock()
    await state.publish_preparing_for_shutdown(mqtt_client=None)
    mqtt_client_mock.publish.assert_not_called()
    assert len(caplog.records) == 1
    assert caplog.records[0].levelno == logging.ERROR
    assert (
        caplog.records[0].message
        == "failed to read logind's PreparingForShutdown property: [error] ('mocked',)"
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("topic_prefix", ["systemctl/hostname", "hostname/systemctl"])
@pytest.mark.parametrize("discovery_prefix", ["homeassistant", "home/assistant"])
@pytest.mark.parametrize("object_id", ["raspberrypi", "debian21"])
@pytest.mark.parametrize("hostname", ["hostname", "host-name"])
@pytest.mark.parametrize(
    ("monitored_system_unit_names", "controlled_system_unit_names"),
    [
        ([], []),
        (
            ["foo.service", "bar.service"],
            ["foo-control.service", "bar-control.service"],
        ),
    ],
)
async def test_publish_homeassistant_device_config(
    # pylint: disable=too-many-arguments,too-many-positional-arguments
    topic_prefix: str,
    discovery_prefix: str,
    object_id: str,
    hostname: str,
    monitored_system_unit_names: typing.List[str],
    controlled_system_unit_names: typing.List[str],
) -> None:
    with unittest.mock.patch("jeepney.io.blocking.open_dbus_connection"):
        state = systemctl_mqtt._State(
            mqtt_topic_prefix=topic_prefix,
            homeassistant_discovery_prefix=discovery_prefix,
            homeassistant_discovery_object_id=object_id,
            poweroff_delay=datetime.timedelta(),
            monitored_system_unit_names=monitored_system_unit_names,
            controlled_system_unit_names=controlled_system_unit_names,
        )
    assert state.monitored_system_unit_names == monitored_system_unit_names
    assert state.controlled_system_unit_names == controlled_system_unit_names
    mqtt_client = unittest.mock.AsyncMock()
    with unittest.mock.patch(
        "systemctl_mqtt._utils.get_hostname", return_value=hostname
    ):
        await state.publish_homeassistant_device_config(mqtt_client=mqtt_client)
    mqtt_client.publish.assert_called_once()
    publish_args, publish_kwargs = mqtt_client.publish.call_args
    assert not publish_args
    assert not publish_kwargs["retain"]
    assert (
        publish_kwargs["topic"] == discovery_prefix + "/device/" + object_id + "/config"
    )
    config = json.loads(publish_kwargs["payload"])
    assert re.match(r"\d+\.\d+\.", config["origin"].pop("sw_version"))
    assert config == {
        "origin": {
            "name": "systemctl-mqtt",
            "support_url": "https://github.com/fphammerle/systemctl-mqtt",
        },
        "device": {"identifiers": [hostname], "name": hostname},
        "availability": {"topic": topic_prefix + "/status"},
        "components": {
            "logind/preparing-for-shutdown": {
                "unique_id": f"systemctl-mqtt-{hostname}-logind-preparing-for-shutdown",
                "object_id": f"{hostname}_logind_preparing_for_shutdown",
                "name": "preparing for shutdown",
                "platform": "binary_sensor",
                "state_topic": topic_prefix + "/preparing-for-shutdown",
                "payload_on": "true",
                "payload_off": "false",
            },
            "logind/poweroff": {
                "unique_id": f"systemctl-mqtt-{hostname}-logind-poweroff",
                "object_id": f"{hostname}_logind_poweroff",
                "name": "poweroff",
                "platform": "button",
                "command_topic": f"{topic_prefix}/poweroff",
            },
            "logind/lock-all-sessions": {
                "unique_id": f"systemctl-mqtt-{hostname}-logind-lock-all-sessions",
                "object_id": f"{hostname}_logind_lock_all_sessions",
                "name": "lock all sessions",
                "platform": "button",
                "command_topic": f"{topic_prefix}/lock-all-sessions",
            },
            "logind/suspend": {
                "unique_id": f"systemctl-mqtt-{hostname}-logind-suspend",
                "object_id": f"{hostname}_logind_suspend",
                "name": "suspend",
                "platform": "button",
                "command_topic": f"{topic_prefix}/suspend",
            },
        }
        | {
            f"unit/system/{n}/active-state": {
                "unique_id": f"systemctl-mqtt-{hostname}-unit-system-{n}-active-state",
                "object_id": f"{hostname}_unit_system_{n}_active_state",
                "name": f"{n} active state",
                "platform": "sensor",
                "state_topic": f"{topic_prefix}/unit/system/{n}/active-state",
            }
            for n in monitored_system_unit_names
        }
        | {
            f"unit/system/{n}/{action}": {
                "unique_id": f"systemctl-mqtt-{hostname}-unit-system-{n}-{action}",
                "object_id": f"{hostname}_unit_system_{n}_{action}",
                "name": f"{n} {action}",
                "platform": "button",
                "command_topic": f"{topic_prefix}/unit/system/{n}/{action}",
            }
            for n in controlled_system_unit_names
            for action in ["restart", "start", "stop"]
        },
    }



import datetime
import unittest.mock

import pytest

import systemctl_mqtt

# pylint: disable=protected-access


@pytest.mark.parametrize(
    "delay", [datetime.timedelta(seconds=4), datetime.timedelta(hours=21)]
)
def test_poweroff_trigger(delay):
    action = systemctl_mqtt._MQTTActionSchedulePoweroff()
    with unittest.mock.patch(
        "systemctl_mqtt._dbus.login_manager.get_login_manager_proxy"
    ), unittest.mock.patch(
        "systemctl_mqtt._dbus.login_manager.schedule_shutdown"
    ) as schedule_shutdown_mock:
        action.trigger(
            state=systemctl_mqtt._State(
                mqtt_topic_prefix="systemctl/hostname",
                homeassistant_discovery_prefix="homeassistant",
                homeassistant_discovery_object_id="node",
                poweroff_delay=delay,
                monitored_system_unit_names=[],
                controlled_system_unit_names=[],
            )
        )
    schedule_shutdown_mock.assert_called_once_with(action="poweroff", delay=delay)


@pytest.mark.parametrize(
    ("topic_suffix", "expected_action_arg"), [("poweroff", "poweroff")]
)
def test_mqtt_topic_suffix_action_mapping_poweroff(topic_suffix, expected_action_arg):
    mqtt_action = systemctl_mqtt._MQTT_TOPIC_SUFFIX_ACTION_MAPPING[topic_suffix]
    login_manager_mock = unittest.mock.MagicMock()
    with unittest.mock.patch(
        "systemctl_mqtt._dbus.login_manager.get_login_manager_proxy",
        return_value=login_manager_mock,
    ):
        mqtt_action.trigger(
            state=systemctl_mqtt._State(
                mqtt_topic_prefix="systemctl/hostname",
                homeassistant_discovery_prefix="homeassistant",
                homeassistant_discovery_object_id="node",
                poweroff_delay=datetime.timedelta(),
                monitored_system_unit_names=[],
                controlled_system_unit_names=[],
            )
        )
    login_manager_mock.ScheduleShutdown.assert_called_once()
    schedule_args, schedule_kwargs = login_manager_mock.ScheduleShutdown.call_args
    assert not schedule_args
    assert schedule_kwargs.pop("action") == expected_action_arg
    assert abs(
        datetime.datetime.now() - schedule_kwargs.pop("time")
    ) < datetime.timedelta(seconds=2)
    assert not schedule_kwargs


def test_mqtt_topic_suffix_action_mapping_lock():
    mqtt_action = systemctl_mqtt._MQTT_TOPIC_SUFFIX_ACTION_MAPPING["lock-all-sessions"]
    login_manager_mock = unittest.mock.MagicMock()
    with unittest.mock.patch(
        "systemctl_mqtt._dbus.login_manager.get_login_manager_proxy",
        return_value=login_manager_mock,
    ):
        mqtt_action.trigger(state="dummy")
    login_manager_mock.LockSessions.assert_called_once_with()


def test_mqtt_topic_suffix_action_mapping_suspend():
    mqtt_action = systemctl_mqtt._MQTT_TOPIC_SUFFIX_ACTION_MAPPING["suspend"]
    login_manager_mock = unittest.mock.MagicMock()
    with unittest.mock.patch(
        "systemctl_mqtt._dbus.login_manager.get_login_manager_proxy",
        return_value=login_manager_mock,
    ):
        mqtt_action.trigger(state="dummy")
    login_manager_mock.Suspend.assert_called_once_with(interactive=False)


def test_poweroff_str():
    assert (
        str(systemctl_mqtt._MQTTActionSchedulePoweroff())
        == "_MQTTActionSchedulePoweroff"
    )



# systemctl-mqtt - MQTT client triggering & reporting shutdown on systemd-based systems
#
# Copyright (C) 2020 Fabian Peter Hammerle <fabian@hammerle.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import re

import systemctl_mqtt._utils

NODE_ID_ALLOWED_CHARS = r"a-zA-Z0-9_-"


def get_default_discovery_object_id() -> str:
    return re.sub(
        f"[^{NODE_ID_ALLOWED_CHARS}]",
        "",
        # pylint: disable=protected-access
        "systemctl-mqtt-" + systemctl_mqtt._utils.get_hostname(),
    )


def validate_discovery_object_id(object_id: str) -> bool:
    return re.match(f"^[{NODE_ID_ALLOWED_CHARS}]+$", object_id) is not None



# systemctl-mqtt - MQTT client triggering & reporting shutdown on systemd-based systems
#
# Copyright (C) 2024 Fabian Peter Hammerle <fabian@hammerle.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import abc

import jeepney


class Properties(jeepney.MessageGenerator):
    """
    https://dbus.freedesktop.org/doc/dbus-specification.html#standard-interfaces-properties
    """

    # pylint: disable=too-few-public-methods

    interface = "org.freedesktop.DBus.Properties"  # overwritten

    # pylint: disable=invalid-name

    def Get(self, property_name: str) -> jeepney.low_level.Message:
        return jeepney.new_method_call(
            remote_obj=jeepney.DBusAddress(
                object_path=self.object_path,
                bus_name=self.bus_name,
                interface="org.freedesktop.DBus.Properties",
            ),
            method="Get",
            signature="ss",
            body=(self.interface, property_name),
        )



# systemctl-mqtt - MQTT client triggering & reporting shutdown on systemd-based systems
#
# Copyright (C) 2024 Fabian Peter Hammerle <fabian@hammerle.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import typing
import unittest.mock

import pytest
import jeepney.io.asyncio
import jeepney.low_level

import systemctl_mqtt

# pylint: disable=protected-access


class DBusErrorResponseMock(jeepney.wrappers.DBusErrorResponse):
    # pylint: disable=missing-class-docstring,super-init-not-called
    def __init__(self, name: str, data: typing.Any):
        self.name = name
        self.data = data


@pytest.mark.asyncio
async def test__get_unit_path() -> None:
    router_mock = unittest.mock.AsyncMock()
    reply_mock = unittest.mock.MagicMock()
    expected_path = "/org/freedesktop/systemd1/unit/ssh_2eservice"
    reply_mock.body = (expected_path,)
    router_mock.send_and_get_reply.return_value = reply_mock
    service_manager = jeepney.io.asyncio.Proxy(
        msggen=systemctl_mqtt._dbus.service_manager.ServiceManager(),
        router=router_mock,
    )
    assert (
        await systemctl_mqtt._get_unit_path(
            service_manager=service_manager, unit_name="ssh.service"
        )
        == expected_path
    )
    router_mock.send_and_get_reply.assert_awaited_once()
    (msg,), send_kwargs = router_mock.send_and_get_reply.await_args
    assert isinstance(msg, jeepney.low_level.Message)
    assert msg.header.fields == {
        jeepney.low_level.HeaderFields.path: "/org/freedesktop/systemd1",
        jeepney.low_level.HeaderFields.destination: "org.freedesktop.systemd1",
        jeepney.low_level.HeaderFields.interface: "org.freedesktop.systemd1.Manager",
        jeepney.low_level.HeaderFields.member: "GetUnit",
        jeepney.low_level.HeaderFields.signature: "s",
    }
    assert msg.body == ("ssh.service",)
    assert not send_kwargs


@pytest.mark.parametrize(
    "action,method",
    [
        ("start", "StartUnit"),
        ("stop", "StopUnit"),
        ("restart", "RestartUnit"),
    ],
)
def test__unit_proxy(action, method):
    mock_proxy = unittest.mock.MagicMock()
    with unittest.mock.patch(
        "systemctl_mqtt._dbus.service_manager.get_service_manager_proxy",
        return_value=mock_proxy,
    ):
        # call the wrapper function dynamically
        getattr(systemctl_mqtt._dbus.service_manager, f"{action}_unit")("foo.service")
        getattr(mock_proxy, method).assert_called_once_with("foo.service", "replace")


@pytest.mark.parametrize(
    "method",
    [
        "StartUnit",
        "StopUnit",
        "RestartUnit",
    ],
)
def test__unit_method_call(method):
    with unittest.mock.patch(
        "jeepney.new_method_call", return_value=unittest.mock.MagicMock()
    ) as mock_method_call:
        mgr = systemctl_mqtt._dbus.service_manager.ServiceManager()
        getattr(mgr, method)("foo.service", "replace")
        mock_method_call.assert_called_once_with(
            remote_obj=mgr,
            method=method,
            signature="ss",
            body=("foo.service", "replace"),
        )


@pytest.mark.parametrize(
    "action,method",
    [
        ("start", "StartUnit"),
        ("stop", "StopUnit"),
        ("restart", "RestartUnit"),
    ],
)
def test__unit_with_exception(action, method):
    mock_proxy = unittest.mock.MagicMock()
    getattr(mock_proxy, method).side_effect = DBusErrorResponseMock(
        "DBus error", ("mocked",)
    )

    with unittest.mock.patch(
        "systemctl_mqtt._dbus.service_manager.get_service_manager_proxy",
        return_value=mock_proxy,
    ), unittest.mock.patch(
        "systemctl_mqtt._dbus.service_manager._LOGGER"
    ) as mock_logger:
        getattr(systemctl_mqtt._dbus.service_manager, f"{action}_unit")(
            "example.service"
        )
        mock_logger.error.assert_called_once_with(
            f"Failed to {action} unit: %s because %s ",
            "example.service",
            "DBus error",
        )



# systemctl-mqtt - MQTT client triggering & reporting shutdown on systemd-based systems
#
# Copyright (C) 2020 Fabian Peter Hammerle <fabian@hammerle.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime
import getpass
import json
import logging
import typing

import jeepney
import jeepney.io.blocking

import systemctl_mqtt._dbus

_LOGGER = logging.getLogger(__name__)

_LOGIN_MANAGER_OBJECT_PATH = "/org/freedesktop/login1"
_LOGIN_MANAGER_INTERFACE = "org.freedesktop.login1.Manager"


def _get_username() -> typing.Optional[str]:
    try:
        return getpass.getuser()
    except OSError:
        # > Traceback (most recent call last):
        # >   File "/usr/local/lib/python3.13/getpass.py", line 173, in getuser
        # >     return pwd.getpwuid(os.getuid())[0]
        # >            ~~~~~~~~~~~~^^^^^^^^^^^^^
        # > KeyError: 'getpwuid(): uid not found: 100'
        #
        # > The above exception was the direct cause of the following exception:
        # > …
        # > OSError: No username set in the environment
        return None


def _log_interactive_authorization_required(
    *, action_label: str, action_id: str
) -> None:
    _LOGGER.error(
        """failed to %s: interactive authorization required

create %s and insert the following rule:
polkit.addRule(function(action, subject) {
    if(action.id === %s && subject.user === %s) {
        return polkit.Result.YES;
    }
});
""",
        action_label,
        "/etc/polkit-1/rules.d/50-systemctl-mqtt.rules",
        json.dumps(action_id),
        json.dumps(_get_username() or "USERNAME"),
    )


def get_login_manager_signal_match_rule(member: str) -> jeepney.MatchRule:
    return jeepney.MatchRule(
        type="signal",
        interface=_LOGIN_MANAGER_INTERFACE,
        member=member,
        path=_LOGIN_MANAGER_OBJECT_PATH,
    )


class LoginManager(systemctl_mqtt._dbus.Properties):  # pylint: disable=protected-access
    """
    https://freedesktop.org/wiki/Software/systemd/logind/

    $ python3 -m jeepney.bindgen \
        --bus unix:path=/var/run/dbus/system_bus_socket \
        --name org.freedesktop.login1 --path /org/freedesktop/login1
    """

    interface = _LOGIN_MANAGER_INTERFACE

    def __init__(self):
        super().__init__(
            object_path=_LOGIN_MANAGER_OBJECT_PATH, bus_name="org.freedesktop.login1"
        )

    # pylint: disable=invalid-name; inherited method names from Manager object

    def ListInhibitors(self) -> jeepney.low_level.Message:
        return jeepney.new_method_call(remote_obj=self, method="ListInhibitors")

    def LockSessions(self) -> jeepney.low_level.Message:
        return jeepney.new_method_call(remote_obj=self, method="LockSessions")

    def CanPowerOff(self) -> jeepney.low_level.Message:
        return jeepney.new_method_call(remote_obj=self, method="CanPowerOff")

    def ScheduleShutdown(
        self, *, action: str, time: datetime.datetime
    ) -> jeepney.low_level.Message:
        return jeepney.new_method_call(
            remote_obj=self,
            method="ScheduleShutdown",
            signature="st",
            body=(action, int(time.timestamp() * 1e6)),  # (type, usec)
        )

    def Suspend(self, *, interactive: bool) -> jeepney.low_level.Message:
        return jeepney.new_method_call(
            remote_obj=self, method="Suspend", signature="b", body=(interactive,)
        )

    def Inhibit(
        self, *, what: str, who: str, why: str, mode: str
    ) -> jeepney.low_level.Message:
        return jeepney.new_method_call(
            remote_obj=self,
            method="Inhibit",
            signature="ssss",
            body=(what, who, why, mode),
        )


def get_login_manager_proxy() -> jeepney.io.blocking.Proxy:
    # https://jeepney.readthedocs.io/en/latest/integrate.html
    # https://gitlab.com/takluyver/jeepney/-/blob/master/examples/aio_notify.py
    return jeepney.io.blocking.Proxy(
        msggen=LoginManager(),
        connection=jeepney.io.blocking.open_dbus_connection(
            bus="SYSTEM",
            # > dbus-broker[…]: Peer :1.… is being disconnected as it does not
            # . support receiving file descriptors it requested.
            enable_fds=True,
        ),
    )


def _log_shutdown_inhibitors(login_manager_proxy: jeepney.io.blocking.Proxy) -> None:
    if _LOGGER.getEffectiveLevel() > logging.DEBUG:
        return
    found_inhibitor = False
    try:
        # https://www.freedesktop.org/wiki/Software/systemd/inhibit/
        (inhibitors,) = login_manager_proxy.ListInhibitors()
        for what, who, why, mode, uid, pid in inhibitors:
            if "shutdown" in what:
                found_inhibitor = True
                _LOGGER.debug(
                    "detected shutdown inhibitor %s (pid=%u, uid=%u, mode=%s): %s",
                    who,
                    pid,
                    uid,
                    mode,
                    why,
                )
    except jeepney.wrappers.DBusErrorResponse as exc:
        _LOGGER.warning("failed to fetch shutdown inhibitors: %s", exc)
        return
    if not found_inhibitor:
        _LOGGER.debug("no shutdown inhibitor locks found")


def schedule_shutdown(*, action: str, delay: datetime.timedelta) -> None:
    # https://github.com/systemd/systemd/blob/v237/src/systemctl/systemctl.c#L8553
    assert action in ["poweroff", "reboot"], action
    time = datetime.datetime.now() + delay
    # datetime.datetime.isoformat(timespec=) not available in python3.5
    # https://github.com/python/cpython/blob/v3.5.9/Lib/datetime.py#L1552
    _LOGGER.info("scheduling %s for %s", action, time.strftime("%Y-%m-%d %H:%M:%S"))
    login_manager = get_login_manager_proxy()
    try:
        # $ gdbus introspect --system --dest org.freedesktop.login1 \
        #       --object-path /org/freedesktop/login1 | grep -A 1 ScheduleShutdown
        # ScheduleShutdown(in  s arg_0,
        #                  in  t arg_1);
        # $ gdbus call --system --dest org.freedesktop.login1 \
        #       --object-path /org/freedesktop/login1 \
        #       --method org.freedesktop.login1.Manager.ScheduleShutdown \
        #       poweroff "$(date --date=10min +%s)000000"
        # $ dbus-send --type=method_call --print-reply --system --dest=org.freedesktop.login1 \
        #       /org/freedesktop/login1 \
        #       org.freedesktop.login1.Manager.ScheduleShutdown \
        #       string:poweroff "uint64:$(date --date=10min +%s)000000"
        login_manager.ScheduleShutdown(action=action, time=time)
    except jeepney.wrappers.DBusErrorResponse as exc:
        if exc.name == "org.freedesktop.DBus.Error.InteractiveAuthorizationRequired":
            _log_interactive_authorization_required(
                action_label="schedule " + action,
                action_id="org.freedesktop.login1."
                + {"poweroff": "power-off"}.get(action, action),
            )
        else:
            _LOGGER.error("failed to schedule %s: %s", action, exc)
    _log_shutdown_inhibitors(login_manager)


def suspend() -> None:
    _LOGGER.info("suspending system")
    get_login_manager_proxy().Suspend(interactive=False)


def lock_all_sessions() -> None:
    """
    $ loginctl lock-sessions
    """
    _LOGGER.info("instruct all sessions to activate screen locks")
    login_manager = get_login_manager_proxy()
    try:
        login_manager.LockSessions()
    except jeepney.wrappers.DBusErrorResponse as exc:
        if exc.name == "org.freedesktop.DBus.Error.InteractiveAuthorizationRequired":
            _log_interactive_authorization_required(
                action_label="lock all sessions",
                action_id="org.freedesktop.login1.lock-sessions",
            )
        else:
            _LOGGER.error("failed to lock all sessions: %s", exc)



# systemctl-mqtt - MQTT client triggering & reporting shutdown on systemd-based systems
#
# Copyright (C) 2020 Fabian Peter Hammerle <fabian@hammerle.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import datetime
import getpass
import logging
import typing
import unittest.mock

import jeepney
import jeepney.low_level
import jeepney.wrappers
import pytest

import systemctl_mqtt._dbus.login_manager

# pylint: disable=protected-access


def test_get_login_manager_proxy():
    login_manager = systemctl_mqtt._dbus.login_manager.get_login_manager_proxy()
    assert isinstance(login_manager, jeepney.io.blocking.Proxy)
    assert login_manager._msggen.interface == "org.freedesktop.login1.Manager"
    # https://freedesktop.org/wiki/Software/systemd/logind/
    assert login_manager.CanPowerOff() in {("yes",), ("challenge",)}


def test_get_service_manager_proxy():
    service_manager = systemctl_mqtt._dbus.service_manager.get_service_manager_proxy()
    assert isinstance(service_manager, jeepney.io.blocking.Proxy)
    assert service_manager._msggen.interface == "org.freedesktop.systemd1.Manager"


def test__log_shutdown_inhibitors_some(caplog):
    login_manager = unittest.mock.MagicMock()
    login_manager.ListInhibitors.return_value = (
        [
            (
                "shutdown:sleep",
                "Developer",
                "Haven't pushed my commits yet",
                "delay",
                1000,
                1234,
            ),
            ("shutdown", "Editor", "", "Unsafed files open", 0, 42),
        ],
    )
    with caplog.at_level(logging.DEBUG):
        systemctl_mqtt._dbus.login_manager._log_shutdown_inhibitors(login_manager)
    assert len(caplog.records) == 2
    assert caplog.records[0].levelno == logging.DEBUG
    assert (
        caplog.records[0].message
        == "detected shutdown inhibitor Developer (pid=1234, uid=1000, mode=delay): "
        + "Haven't pushed my commits yet"
    )


def test__log_shutdown_inhibitors_none(caplog):
    login_manager = unittest.mock.MagicMock()
    login_manager.ListInhibitors.return_value = ([],)
    with caplog.at_level(logging.DEBUG):
        systemctl_mqtt._dbus.login_manager._log_shutdown_inhibitors(login_manager)
    assert len(caplog.records) == 1
    assert caplog.records[0].levelno == logging.DEBUG
    assert caplog.records[0].message == "no shutdown inhibitor locks found"


def test__log_shutdown_inhibitors_fail(caplog):
    login_manager = unittest.mock.MagicMock()
    login_manager.ListInhibitors.side_effect = DBusErrorResponseMock("error", "mocked")
    with caplog.at_level(logging.DEBUG):
        systemctl_mqtt._dbus.login_manager._log_shutdown_inhibitors(login_manager)
    assert len(caplog.records) == 1
    assert caplog.records[0].levelno == logging.WARNING
    assert (
        caplog.records[0].message
        == "failed to fetch shutdown inhibitors: [error] mocked"
    )


@pytest.mark.parametrize("action", ["poweroff", "reboot"])
@pytest.mark.parametrize("delay", [datetime.timedelta(0), datetime.timedelta(hours=1)])
def test__schedule_shutdown(action, delay):
    login_manager_mock = unittest.mock.MagicMock()
    with unittest.mock.patch(
        "systemctl_mqtt._dbus.login_manager.get_login_manager_proxy",
        return_value=login_manager_mock,
    ):
        login_manager_mock.ListInhibitors.return_value = ([],)
        systemctl_mqtt._dbus.login_manager.schedule_shutdown(action=action, delay=delay)
    login_manager_mock.ScheduleShutdown.assert_called_once()
    schedule_args, schedule_kwargs = login_manager_mock.ScheduleShutdown.call_args
    assert not schedule_args
    assert schedule_kwargs.pop("action") == action
    actual_delay = schedule_kwargs.pop("time") - datetime.datetime.now()
    assert actual_delay.total_seconds() == pytest.approx(delay.total_seconds(), abs=0.1)
    assert not schedule_kwargs


class DBusErrorResponseMock(jeepney.wrappers.DBusErrorResponse):
    # pylint: disable=missing-class-docstring,super-init-not-called
    def __init__(self, name: str, data: typing.Any):
        self.name = name
        self.data = data


@pytest.mark.parametrize(
    ("action", "error_name", "error_message", "log_message"),
    [
        (
            "poweroff",
            "test error",
            "test message",
            "[test error] ('test message',)",
        ),
        (
            "poweroff",
            "org.freedesktop.DBus.Error.InteractiveAuthorizationRequired",
            "Interactive authentication required.",
            """interactive authorization required

create /etc/polkit-1/rules.d/50-systemctl-mqtt.rules and insert the following rule:
polkit.addRule(function(action, subject) {
    if(action.id === "org.freedesktop.login1.power-off" && subject.user === "{{username}}") {
        return polkit.Result.YES;
    }
});
""".replace(
                "{{username}}", getpass.getuser()
            ),
        ),
        (
            "reboot",
            "org.freedesktop.DBus.Error.InteractiveAuthorizationRequired",
            "Interactive authentication required.",
            """interactive authorization required

create /etc/polkit-1/rules.d/50-systemctl-mqtt.rules and insert the following rule:
polkit.addRule(function(action, subject) {
    if(action.id === "org.freedesktop.login1.reboot" && subject.user === "{{username}}") {
        return polkit.Result.YES;
    }
});
""".replace(
                "{{username}}", getpass.getuser()
            ),
        ),
    ],
)
def test__schedule_shutdown_fail(
    caplog, action: str, error_name: str, error_message: str, log_message: str
) -> None:
    login_manager_mock = unittest.mock.MagicMock()
    login_manager_mock.ScheduleShutdown.side_effect = DBusErrorResponseMock(
        name=error_name,
        data=(error_message,),
    )
    login_manager_mock.ListInhibitors.return_value = ([],)
    with unittest.mock.patch(
        "systemctl_mqtt._dbus.login_manager.get_login_manager_proxy",
        return_value=login_manager_mock,
    ), caplog.at_level(logging.DEBUG):
        systemctl_mqtt._dbus.login_manager.schedule_shutdown(
            action=action, delay=datetime.timedelta(seconds=21)
        )
    login_manager_mock.ScheduleShutdown.assert_called_once()
    assert len(caplog.records) == 3
    assert caplog.records[0].levelno == logging.INFO
    assert caplog.records[0].message.startswith(f"scheduling {action} for ")
    assert caplog.records[1].levelno == logging.ERROR
    assert caplog.records[1].message == f"failed to schedule {action}: {log_message}"
    assert "inhibitor" in caplog.records[2].message


@pytest.mark.parametrize("action", ["poweroff"])
@pytest.mark.parametrize(
    ("error_name", "error_message", "log_message"),
    [
        (
            "org.freedesktop.DBus.Error.InteractiveAuthorizationRequired",
            "Interactive authentication required.",
            """interactive authorization required

create /etc/polkit-1/rules.d/50-systemctl-mqtt.rules and insert the following rule:
polkit.addRule(function(action, subject) {
    if(action.id === "org.freedesktop.login1.power-off" && subject.user === "USERNAME") {
        return polkit.Result.YES;
    }
});
""",
        ),
    ],
)
def test__schedule_shutdown_fail_no_username(
    caplog, action, error_name, error_message, log_message
):
    login_manager_mock = unittest.mock.MagicMock()
    login_manager_mock.ScheduleShutdown.side_effect = DBusErrorResponseMock(
        name=error_name,
        data=(error_message,),
    )
    login_manager_mock.ListInhibitors.return_value = ([],)
    with unittest.mock.patch(
        "systemctl_mqtt._dbus.login_manager.get_login_manager_proxy",
        return_value=login_manager_mock,
    ), unittest.mock.patch(
        "getpass.getuser", side_effect=OSError("No username set in the environment")
    ), caplog.at_level(
        logging.ERROR
    ):
        systemctl_mqtt._dbus.login_manager.schedule_shutdown(
            action=action, delay=datetime.timedelta(seconds=21)
        )
    login_manager_mock.ScheduleShutdown.assert_called_once()
    assert len(caplog.records) == 1
    assert caplog.records[0].levelno == logging.ERROR
    assert caplog.records[0].message == f"failed to schedule {action}: {log_message}"


def test_suspend(caplog):
    login_manager_mock = unittest.mock.MagicMock()
    with unittest.mock.patch(
        "systemctl_mqtt._dbus.login_manager.get_login_manager_proxy",
        return_value=login_manager_mock,
    ), caplog.at_level(logging.INFO):
        systemctl_mqtt._dbus.login_manager.suspend()
    login_manager_mock.Suspend.assert_called_once_with(interactive=False)
    assert len(caplog.records) == 1
    assert caplog.records[0].levelno == logging.INFO
    assert caplog.records[0].message == "suspending system"


def test_lock_all_sessions(caplog):
    login_manager_mock = unittest.mock.MagicMock()
    with unittest.mock.patch(
        "systemctl_mqtt._dbus.login_manager.get_login_manager_proxy",
        return_value=login_manager_mock,
    ), caplog.at_level(logging.INFO):
        systemctl_mqtt._dbus.login_manager.lock_all_sessions()
    login_manager_mock.LockSessions.assert_called_once_with()
    assert len(caplog.records) == 1
    assert caplog.records[0].levelno == logging.INFO
    assert caplog.records[0].message == "instruct all sessions to activate screen locks"


@pytest.mark.parametrize(
    ("error_name", "error_message", "log_message"),
    [
        (
            "test error",
            "test message",
            "[test error] ('test message',)",
        ),
        (
            "org.freedesktop.DBus.Error.InteractiveAuthorizationRequired",
            "Interactive authentication required.",
            """interactive authorization required

create /etc/polkit-1/rules.d/50-systemctl-mqtt.rules and insert the following rule:
polkit.addRule(function(action, subject) {
    if(action.id === "org.freedesktop.login1.lock-sessions" && subject.user === "{{username}}") {
        return polkit.Result.YES;
    }
});
""".replace(
                "{{username}}", getpass.getuser()
            ),
        ),
    ],
)
def test_lock_all_sessions_fail(
    caplog: pytest.LogCaptureFixture,
    error_name: str,
    error_message: str,
    log_message: str,
) -> None:
    login_manager_mock = unittest.mock.MagicMock()
    login_manager_mock.LockSessions.side_effect = DBusErrorResponseMock(
        name=error_name, data=(error_message,)
    )
    with unittest.mock.patch(
        "systemctl_mqtt._dbus.login_manager.get_login_manager_proxy",
        return_value=login_manager_mock,
    ), caplog.at_level(logging.ERROR):
        systemctl_mqtt._dbus.login_manager.lock_all_sessions()
    login_manager_mock.LockSessions.assert_called_once()
    assert len(caplog.records) == 1
    assert caplog.records[0].levelno == logging.ERROR
    assert caplog.records[0].message == f"failed to lock all sessions: {log_message}"


async def _get_unit_path_mock(  # pylint: disable=unused-argument
    *, service_manager: jeepney.io.asyncio.Proxy, unit_name: str
) -> str:
    return "/org/freedesktop/systemd1/unit/" + unit_name


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "monitored_system_unit_names", [[], ["foo.service", "bar.service"]]
)
async def test__dbus_signal_loop(monitored_system_unit_names: typing.List[str]) -> None:
    # pylint: disable=too-many-locals,too-many-arguments
    state_mock = unittest.mock.AsyncMock()
    with unittest.mock.patch(
        "jeepney.io.asyncio.open_dbus_router",
    ) as open_dbus_router_mock, unittest.mock.patch(
        "systemctl_mqtt._get_unit_path", _get_unit_path_mock
    ), unittest.mock.patch(
        "systemctl_mqtt._dbus_signal_loop_unit"
    ) as dbus_signal_loop_unit_mock:
        async with open_dbus_router_mock() as dbus_router_mock:
            pass
        add_match_reply = unittest.mock.Mock()
        add_match_reply.body = ()
        dbus_router_mock.send_and_get_reply.return_value = add_match_reply
        msg_queue: asyncio.Queue[jeepney.low_level.Message] = asyncio.Queue()
        await msg_queue.put(jeepney.low_level.Message(header=None, body=(False,)))
        await msg_queue.put(jeepney.low_level.Message(header=None, body=(True,)))
        await msg_queue.put(jeepney.low_level.Message(header=None, body=(False,)))
        dbus_router_mock.filter = unittest.mock.MagicMock()
        dbus_router_mock.filter.return_value.__enter__.return_value = msg_queue
        state_mock.monitored_system_unit_names = monitored_system_unit_names
        # asyncio.TaskGroup added in python3.11
        loop_task = asyncio.create_task(
            systemctl_mqtt._dbus_signal_loop(
                state=state_mock, mqtt_client=unittest.mock.MagicMock()
            )
        )

        async def _abort_after_msg_queue():
            await msg_queue.join()
            loop_task.cancel()

        with pytest.raises(asyncio.exceptions.CancelledError):
            await asyncio.gather(*(loop_task, _abort_after_msg_queue()))
    assert unittest.mock.call(bus="SYSTEM") in open_dbus_router_mock.call_args_list
    dbus_router_mock.filter.assert_called_once()
    (filter_match_rule,) = dbus_router_mock.filter.call_args[0]
    assert (
        filter_match_rule.header_fields["interface"] == "org.freedesktop.login1.Manager"
    )
    assert filter_match_rule.header_fields["member"] == "PrepareForShutdown"
    add_match_msg = dbus_router_mock.send_and_get_reply.call_args[0][0]
    assert (
        add_match_msg.header.fields[jeepney.low_level.HeaderFields.member] == "AddMatch"
    )
    assert add_match_msg.body == (
        "interface='org.freedesktop.login1.Manager',member='PrepareForShutdown'"
        ",path='/org/freedesktop/login1',type='signal'",
    )
    assert [
        c[1]["active"] for c in state_mock.preparing_for_shutdown_handler.call_args_list
    ] == [False, True, False]
    assert not any(args for args, _ in dbus_signal_loop_unit_mock.await_args_list)
    dbus_signal_loop_unit_kwargs = [
        kwargs for _, kwargs in dbus_signal_loop_unit_mock.await_args_list
    ]
    assert [(a["unit_name"], a["unit_path"]) for a in dbus_signal_loop_unit_kwargs] == [
        (n, f"/org/freedesktop/systemd1/unit/{n}") for n in monitored_system_unit_names
    ]


def _mock_get_active_state_reply(state: str) -> unittest.mock.MagicMock:
    reply_mock = unittest.mock.MagicMock()
    reply_mock.body = (("s", state),)
    return reply_mock


@pytest.mark.asyncio
async def test__dbus_signal_loop_unit() -> None:
    state = systemctl_mqtt._State(
        mqtt_topic_prefix="prefix",
        homeassistant_discovery_prefix="unused",
        homeassistant_discovery_object_id="unused",
        poweroff_delay=datetime.timedelta(),
        monitored_system_unit_names=[],
        controlled_system_unit_names=[],
    )
    mqtt_client_mock = unittest.mock.AsyncMock()
    dbus_router_mock = unittest.mock.AsyncMock()
    bus_proxy_mock = unittest.mock.AsyncMock()
    bus_proxy_mock.AddMatch.return_value = ()
    get_active_state_reply_mock = unittest.mock.MagicMock()
    get_active_state_reply_mock.body = (("s", "active"),)
    states = [
        "active",
        "deactivating",
        "inactive",
        "inactive",
        "activating",
        "active",
        "active",
        "active",
        "inactive",
    ]
    dbus_router_mock.send_and_get_reply.side_effect = [
        _mock_get_active_state_reply(s) for s in states
    ]
    msg_queue: asyncio.Queue[jeepney.low_level.Message] = asyncio.Queue()
    for _ in range(len(states) - 1):
        await msg_queue.put(jeepney.low_level.Message(header=None, body=()))
    dbus_router_mock.filter = unittest.mock.MagicMock()
    dbus_router_mock.filter.return_value.__enter__.return_value = msg_queue
    loop_task = asyncio.create_task(
        systemctl_mqtt._dbus_signal_loop_unit(
            state=state,
            mqtt_client=mqtt_client_mock,
            dbus_router=dbus_router_mock,
            bus_proxy=bus_proxy_mock,
            unit_name="foo.service",
            unit_path="/org/freedesktop/systemd1/unit/whatever_2eservice",
        )
    )

    async def _abort_after_msg_queue():
        await msg_queue.join()
        loop_task.cancel()

    with pytest.raises(asyncio.exceptions.CancelledError):
        await asyncio.gather(*(loop_task, _abort_after_msg_queue()))
    bus_proxy_mock.AddMatch.assert_awaited_once()
    ((match_rule,), add_match_kwargs) = bus_proxy_mock.AddMatch.await_args
    assert match_rule.header_fields["interface"] == "org.freedesktop.DBus.Properties"
    assert match_rule.header_fields["member"] == "PropertiesChanged"
    assert not add_match_kwargs
    assert mqtt_client_mock.publish.await_args_list == [
        unittest.mock.call(
            topic="prefix/unit/system/foo.service/active-state", payload=s
        )
        for s in [  # consecutive duplicates filtered
            "active",
            "deactivating",
            "inactive",
            "activating",
            "active",
            "inactive",
        ]
    ]



# systemctl-mqtt - MQTT client triggering & reporting shutdown on systemd-based systems
#
# Copyright (C) 2020 Fabian Peter Hammerle <fabian@hammerle.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime
import logging
import typing
import unittest.mock

import pytest

import systemctl_mqtt
import systemctl_mqtt._homeassistant
import systemctl_mqtt._utils

# pylint: disable=protected-access,too-many-positional-arguments


@pytest.mark.parametrize(
    ("args", "log_level"),
    [
        ([], logging.INFO),
        (["--log-level", "debug"], logging.DEBUG),
        (["--log-level", "info"], logging.INFO),
        (["--log-level", "warning"], logging.WARNING),
        (["--log-level", "error"], logging.ERROR),
        (["--log-level", "critical"], logging.CRITICAL),
    ],
)
def test__main_log_level(args: typing.List[str], log_level: int) -> None:
    with unittest.mock.patch("systemctl_mqtt._run") as run_mock, unittest.mock.patch(
        "sys.argv", ["", "--mqtt-host", "mqtt-broker.local"] + args
    ):
        systemctl_mqtt._main()
    run_mock.assert_called_once()
    assert logging.root.getEffectiveLevel() == log_level


@pytest.mark.parametrize(
    (
        "argv",
        "expected_mqtt_host",
        "expected_mqtt_port",
        "expected_mqtt_disable_tls",
        "expected_username",
        "expected_password",
        "expected_topic_prefix",
    ),
    [
        (
            ["", "--mqtt-host", "mqtt-broker.local"],
            "mqtt-broker.local",
            8883,
            False,
            None,
            None,
            None,
        ),
        (
            ["", "--mqtt-host", "mqtt-broker.local", "--mqtt-disable-tls"],
            "mqtt-broker.local",
            1883,
            True,
            None,
            None,
            None,
        ),
        (
            ["", "--mqtt-host", "mqtt-broker.local", "--mqtt-port", "8883"],
            "mqtt-broker.local",
            8883,
            False,
            None,
            None,
            None,
        ),
        (
            ["", "--mqtt-host", "mqtt-broker.local", "--mqtt-port", "8884"],
            "mqtt-broker.local",
            8884,
            False,
            None,
            None,
            None,
        ),
        (
            [
                "",
                "--mqtt-host",
                "mqtt-broker.local",
                "--mqtt-port",
                "8884",
                "--mqtt-disable-tls",
            ],
            "mqtt-broker.local",
            8884,
            True,
            None,
            None,
            None,
        ),
        (
            ["", "--mqtt-host", "mqtt-broker.local", "--mqtt-username", "me"],
            "mqtt-broker.local",
            8883,
            False,
            "me",
            None,
            None,
        ),
        (
            [
                "",
                "--mqtt-host",
                "mqtt-broker.local",
                "--mqtt-username",
                "me",
                "--mqtt-password",
                "secret",
            ],
            "mqtt-broker.local",
            8883,
            False,
            "me",
            "secret",
            None,
        ),
        (
            [
                "",
                "--mqtt-host",
                "mqtt-broker.local",
                "--mqtt-topic-prefix",
                "system/command",
            ],
            "mqtt-broker.local",
            8883,
            False,
            None,
            None,
            "system/command",
        ),
    ],
)
def test__main(
    argv,
    expected_mqtt_host,
    expected_mqtt_port,
    expected_mqtt_disable_tls,
    expected_username,
    expected_password,
    expected_topic_prefix: typing.Optional[str],
):
    # pylint: disable=too-many-arguments
    with unittest.mock.patch("systemctl_mqtt._run") as run_mock, unittest.mock.patch(
        "sys.argv", argv
    ), unittest.mock.patch(
        "systemctl_mqtt._utils.get_hostname", return_value="hostname"
    ):
        # pylint: disable=protected-access
        systemctl_mqtt._main()
    run_mock.assert_called_once_with(
        mqtt_host=expected_mqtt_host,
        mqtt_port=expected_mqtt_port,
        mqtt_disable_tls=expected_mqtt_disable_tls,
        mqtt_username=expected_username,
        mqtt_password=expected_password,
        mqtt_topic_prefix=expected_topic_prefix or "systemctl/hostname",
        homeassistant_discovery_prefix="homeassistant",
        homeassistant_discovery_object_id="systemctl-mqtt-hostname",
        poweroff_delay=datetime.timedelta(seconds=4),
        monitored_system_unit_names=[],
        controlled_system_unit_names=[],
    )


@pytest.mark.parametrize(
    ("password_file_content", "expected_password"),
    [
        ("secret", "secret"),
        ("secret space", "secret space"),
        ("secret   ", "secret   "),
        ("  secret ", "  secret "),
        ("secret\n", "secret"),
        ("secret\n\n", "secret\n"),
        ("secret\r\n", "secret"),
        ("secret\n\r\n", "secret\n"),
        ("你好\n", "你好"),
    ],
)
def test__main_password_file(tmpdir, password_file_content, expected_password):
    mqtt_password_path = tmpdir.join("mqtt-password")
    with mqtt_password_path.open("w") as mqtt_password_file:
        mqtt_password_file.write(password_file_content)
    with unittest.mock.patch("systemctl_mqtt._run") as run_mock, unittest.mock.patch(
        "sys.argv",
        [
            "",
            "--mqtt-host",
            "localhost",
            "--mqtt-username",
            "me",
            "--mqtt-password-file",
            str(mqtt_password_path),
        ],
    ), unittest.mock.patch(
        "systemctl_mqtt._utils.get_hostname", return_value="hostname"
    ):
        # pylint: disable=protected-access
        systemctl_mqtt._main()
    run_mock.assert_called_once_with(
        mqtt_host="localhost",
        mqtt_port=8883,
        mqtt_disable_tls=False,
        mqtt_username="me",
        mqtt_password=expected_password,
        mqtt_topic_prefix="systemctl/hostname",
        homeassistant_discovery_prefix="homeassistant",
        homeassistant_discovery_object_id="systemctl-mqtt-hostname",
        poweroff_delay=datetime.timedelta(seconds=4),
        monitored_system_unit_names=[],
        controlled_system_unit_names=[],
    )


def test__main_password_file_collision(capsys):
    with unittest.mock.patch(
        "sys.argv",
        [
            "",
            "--mqtt-host",
            "localhost",
            "--mqtt-username",
            "me",
            "--mqtt-password",
            "secret",
            "--mqtt-password-file",
            "/var/lib/secrets/mqtt/password",
        ],
    ):
        with pytest.raises(SystemExit):
            # pylint: disable=protected-access
            systemctl_mqtt._main()
    out, err = capsys.readouterr()
    assert not out
    assert (
        "argument --mqtt-password-file: not allowed with argument --mqtt-password\n"
        in err
    )


@pytest.mark.parametrize(
    ("args", "discovery_prefix"),
    [
        ([], "homeassistant"),
        (["--homeassistant-discovery-prefix", "home/assistant"], "home/assistant"),
    ],
)
def test__main_homeassistant_discovery_prefix(args, discovery_prefix):
    with unittest.mock.patch("systemctl_mqtt._run") as run_mock, unittest.mock.patch(
        "sys.argv", ["", "--mqtt-host", "mqtt-broker.local"] + args
    ):
        systemctl_mqtt._main()
    run_mock.assert_called_once()
    assert run_mock.call_args[1]["homeassistant_discovery_prefix"] == discovery_prefix


@pytest.mark.parametrize(
    ("args", "object_id"),
    [
        ([], "systemctl-mqtt-fallback"),
        (["--homeassistant-discovery-object-id", "raspberrypi"], "raspberrypi"),
    ],
)
def test__main_homeassistant_discovery_object_id(args, object_id):
    with unittest.mock.patch("systemctl_mqtt._run") as run_mock, unittest.mock.patch(
        "sys.argv", ["", "--mqtt-host", "mqtt-broker.local"] + args
    ), unittest.mock.patch(
        "systemctl_mqtt._utils.get_hostname", return_value="fallback"
    ):
        systemctl_mqtt._main()
    run_mock.assert_called_once()
    assert run_mock.call_args[1]["homeassistant_discovery_object_id"] == object_id


@pytest.mark.parametrize(
    "args",
    [
        ["--homeassistant-discovery-object-id", "no pe"],
        ["--homeassistant-discovery-object-id", ""],
    ],
)
def test__main_homeassistant_discovery_object_id_invalid(args):
    with unittest.mock.patch(
        "sys.argv", ["", "--mqtt-host", "mqtt-broker.local"] + args
    ):
        with pytest.raises(ValueError):
            systemctl_mqtt._main()


@pytest.mark.parametrize(
    ("args", "poweroff_delay"),
    [
        ([], datetime.timedelta(seconds=4)),
        (["--poweroff-delay-seconds", "42.21"], datetime.timedelta(seconds=42.21)),
        (["--poweroff-delay-seconds", "3600"], datetime.timedelta(hours=1)),
    ],
)
def test__main_poweroff_delay(args, poweroff_delay):
    with unittest.mock.patch("systemctl_mqtt._run") as run_mock, unittest.mock.patch(
        "sys.argv", ["", "--mqtt-host", "mqtt-broker.local"] + args
    ):
        systemctl_mqtt._main()
    run_mock.assert_called_once()
    assert run_mock.call_args[1]["poweroff_delay"] == poweroff_delay



# systemctl-mqtt - MQTT client triggering & reporting shutdown on systemd-based systems
#
# Copyright (C) 2020 Fabian Peter Hammerle <fabian@hammerle.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import abc
import argparse
import asyncio
import datetime
import functools
import importlib.metadata
import json
import logging
import os
import pathlib
import socket
import ssl
import threading
import typing

import aiomqtt
import jeepney
import jeepney.bus_messages
import jeepney.io.asyncio

import systemctl_mqtt._dbus.login_manager
import systemctl_mqtt._dbus.service_manager
import systemctl_mqtt._homeassistant
import systemctl_mqtt._mqtt

_MQTT_DEFAULT_PORT = 1883
_MQTT_DEFAULT_TLS_PORT = 8883
# > payload_not_available string (Optional, default: offline)
# https://web.archive.org/web/20250101075341/https://www.home-assistant.io/integrations/sensor.mqtt/#payload_not_available
_MQTT_PAYLOAD_NOT_AVAILABLE = "offline"
_MQTT_PAYLOAD_AVAILABLE = "online"
_ARGUMENT_LOG_LEVEL_MAPPING = {
    a: getattr(logging, a.upper())
    for a in ("debug", "info", "warning", "error", "critical")
}

_LOGGER = logging.getLogger(__name__)


class _State:
    # pylint: disable=too-many-instance-attributes
    def __init__(  # pylint: disable=too-many-arguments
        self,
        *,
        mqtt_topic_prefix: str,
        homeassistant_discovery_prefix: str,
        homeassistant_discovery_object_id: str,
        poweroff_delay: datetime.timedelta,
        monitored_system_unit_names: typing.List[str],
        controlled_system_unit_names: typing.List[str],
    ) -> None:
        self._mqtt_topic_prefix = mqtt_topic_prefix
        self._homeassistant_discovery_prefix = homeassistant_discovery_prefix
        self._homeassistant_discovery_object_id = homeassistant_discovery_object_id
        self._login_manager = (
            systemctl_mqtt._dbus.login_manager.get_login_manager_proxy()
        )
        self._shutdown_lock: typing.Optional[jeepney.fds.FileDescriptor] = None
        self._shutdown_lock_mutex = threading.Lock()
        self.poweroff_delay = poweroff_delay
        self._monitored_system_unit_names = monitored_system_unit_names
        self._controlled_system_unit_names = controlled_system_unit_names

    @property
    def mqtt_topic_prefix(self) -> str:
        return self._mqtt_topic_prefix

    @property
    def mqtt_availability_topic(self) -> str:
        # > mqtt.ATTR_TOPIC: "homeassistant/status",
        # https://github.com/home-assistant/core/blob/2024.12.5/tests/components/mqtt/conftest.py#L23
        # > _MQTT_AVAILABILITY_TOPIC = "switchbot-mqtt/status"
        # https://github.com/fphammerle/switchbot-mqtt/blob/v3.3.1/switchbot_mqtt/__init__.py#L30
        return self._mqtt_topic_prefix + "/status"

    def get_system_unit_active_state_mqtt_topic(self, *, unit_name: str) -> str:
        return self._mqtt_topic_prefix + "/unit/system/" + unit_name + "/active-state"

    def get_system_unit_start_mqtt_topic(self, *, unit_name: str) -> str:
        return self._mqtt_topic_prefix + "/unit/system/" + unit_name + "/start"

    def get_system_unit_stop_mqtt_topic(self, *, unit_name: str) -> str:
        return self._mqtt_topic_prefix + "/unit/system/" + unit_name + "/stop"

    def get_system_unit_restart_mqtt_topic(self, *, unit_name: str) -> str:
        return self._mqtt_topic_prefix + "/unit/system/" + unit_name + "/restart"

    @property
    def monitored_system_unit_names(self) -> typing.List[str]:
        return self._monitored_system_unit_names

    @property
    def controlled_system_unit_names(self) -> typing.List[str]:
        return self._controlled_system_unit_names

    @property
    def shutdown_lock_acquired(self) -> bool:
        return self._shutdown_lock is not None

    def acquire_shutdown_lock(self) -> None:
        with self._shutdown_lock_mutex:
            assert self._shutdown_lock is None
            # https://www.freedesktop.org/wiki/Software/systemd/inhibit/
            (self._shutdown_lock,) = self._login_manager.Inhibit(
                what="shutdown",
                who="systemctl-mqtt",
                why="Report shutdown via MQTT",
                mode="delay",
            )
            assert isinstance(
                self._shutdown_lock, jeepney.fds.FileDescriptor
            ), self._shutdown_lock
            _LOGGER.debug("acquired shutdown inhibitor lock")

    def release_shutdown_lock(self) -> None:
        with self._shutdown_lock_mutex:
            if self._shutdown_lock:
                self._shutdown_lock.close()
                _LOGGER.debug("released shutdown inhibitor lock")
                self._shutdown_lock = None

    @property
    def _preparing_for_shutdown_topic(self) -> str:
        return self.mqtt_topic_prefix + "/preparing-for-shutdown"

    async def _publish_preparing_for_shutdown(
        self, *, mqtt_client: aiomqtt.Client, active: bool
    ) -> None:
        topic = self._preparing_for_shutdown_topic
        # pylint: disable=protected-access
        payload = systemctl_mqtt._mqtt.encode_bool(active)
        _LOGGER.info("publishing %r on %s", payload, topic)
        await mqtt_client.publish(topic=topic, payload=payload, retain=False)

    async def preparing_for_shutdown_handler(
        self, active: bool, mqtt_client: aiomqtt.Client
    ) -> None:
        active = bool(active)
        await self._publish_preparing_for_shutdown(
            mqtt_client=mqtt_client, active=active
        )
        if active:
            self.release_shutdown_lock()
        else:
            self.acquire_shutdown_lock()

    async def publish_preparing_for_shutdown(self, mqtt_client: aiomqtt.Client) -> None:
        try:
            ((return_type, active),) = self._login_manager.Get("PreparingForShutdown")
        except jeepney.wrappers.DBusErrorResponse as exc:
            _LOGGER.error(
                "failed to read logind's PreparingForShutdown property: %s", exc
            )
            return
        assert return_type == "b", return_type
        assert isinstance(active, bool), active
        await self._publish_preparing_for_shutdown(
            mqtt_client=mqtt_client, active=active
        )

    async def publish_homeassistant_device_config(
        self, mqtt_client: aiomqtt.Client
    ) -> None:
        # <discovery_prefix>/<component>/[<node_id>/]<object_id>/config
        # https://www.home-assistant.io/integrations/mqtt/#mqtt-discovery
        discovery_topic = "/".join(
            (
                self._homeassistant_discovery_prefix,
                "device",
                self._homeassistant_discovery_object_id,
                "config",
            )
        )
        hostname = (
            # pylint: disable=protected-access; function in internal module
            systemctl_mqtt._utils.get_hostname()
        )
        package_metadata = importlib.metadata.metadata(__name__)
        unique_id_prefix = "systemctl-mqtt-" + hostname
        config = {
            "device": {"identifiers": [hostname], "name": hostname},
            "origin": {
                "name": package_metadata["Name"],
                "sw_version": package_metadata["Version"],
                "support_url": package_metadata["Home-page"],
            },
            "availability": {"topic": self.mqtt_availability_topic},
            "components": {
                "logind/preparing-for-shutdown": {
                    "unique_id": unique_id_prefix + "-logind-preparing-for-shutdown",
                    "object_id": f"{hostname}_logind_preparing_for_shutdown",  # entity id
                    "name": "preparing for shutdown",  # home assistant prepends device name
                    "platform": "binary_sensor",
                    "state_topic": self._preparing_for_shutdown_topic,
                    # pylint: disable=protected-access
                    "payload_on": systemctl_mqtt._mqtt.encode_bool(True),
                    "payload_off": systemctl_mqtt._mqtt.encode_bool(False),
                },
            },
        }
        for mqtt_topic_suffix in _MQTT_TOPIC_SUFFIX_ACTION_MAPPING.keys():
            # false positive warning by mypy:
            # > Unsupported target for indexed assignment
            config["components"]["logind/" + mqtt_topic_suffix] = {  # type: ignore
                "unique_id": unique_id_prefix + "-logind-" + mqtt_topic_suffix,
                "object_id": hostname
                + "_logind_"
                + mqtt_topic_suffix.replace("-", "_"),  # entity id
                "name": mqtt_topic_suffix.replace("-", " "),
                "platform": "button",
                "command_topic": self.mqtt_topic_prefix + "/" + mqtt_topic_suffix,
            }
        for unit_name in self._monitored_system_unit_names:
            config["components"]["unit/system/" + unit_name + "/active-state"] = {  # type: ignore
                "unique_id": f"{unique_id_prefix}-unit-system-{unit_name}-active-state",
                "object_id": f"{hostname}_unit_system_{unit_name}_active_state",
                "name": f"{unit_name} active state",
                "platform": "sensor",
                "state_topic": self.get_system_unit_active_state_mqtt_topic(
                    unit_name=unit_name
                ),
            }
        for unit_name in self._controlled_system_unit_names:
            config["components"]["unit/system/" + unit_name + "/start"] = {  # type: ignore
                "unique_id": f"{unique_id_prefix}-unit-system-{unit_name}-start",
                "object_id": f"{hostname}_unit_system_{unit_name}_start",
                "name": f"{unit_name} start",
                "platform": "button",
                "command_topic": self.get_system_unit_start_mqtt_topic(
                    unit_name=unit_name
                ),
            }
            config["components"]["unit/system/" + unit_name + "/stop"] = {  # type: ignore
                "unique_id": f"{unique_id_prefix}-unit-system-{unit_name}-stop",
                "object_id": f"{hostname}_unit_system_{unit_name}_stop",
                "name": f"{unit_name} stop",
                "platform": "button",
                "command_topic": self.get_system_unit_stop_mqtt_topic(
                    unit_name=unit_name
                ),
            }
            config["components"]["unit/system/" + unit_name + "/restart"] = {  # type: ignore
                "unique_id": f"{unique_id_prefix}-unit-system-{unit_name}-restart",
                "object_id": f"{hostname}_unit_system_{unit_name}_restart",
                "name": f"{unit_name} restart",
                "platform": "button",
                "command_topic": self.get_system_unit_restart_mqtt_topic(
                    unit_name=unit_name
                ),
            }
        _LOGGER.debug("publishing home assistant config on %s", discovery_topic)
        await mqtt_client.publish(
            topic=discovery_topic, payload=json.dumps(config), retain=False
        )


class _MQTTAction(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def trigger(self, state: _State) -> None:
        pass  # pragma: no cover

    def __str__(self) -> str:
        return type(self).__name__


class _MQTTActionSchedulePoweroff(_MQTTAction):
    # pylint: disable=too-few-public-methods
    def trigger(self, state: _State) -> None:
        # pylint: disable=protected-access
        systemctl_mqtt._dbus.login_manager.schedule_shutdown(
            action="poweroff", delay=state.poweroff_delay
        )


class _MQTTActionStartUnit(_MQTTAction):
    # pylint: disable=protected-access,too-few-public-methods
    def __init__(self, unit_name: str):
        self._unit_name = unit_name

    def trigger(self, state: _State) -> None:
        systemctl_mqtt._dbus.service_manager.start_unit(unit_name=self._unit_name)


class _MQTTActionStopUnit(_MQTTAction):
    # pylint: disable=protected-access,too-few-public-methods
    def __init__(self, unit_name: str):
        self._unit_name = unit_name

    def trigger(self, state: _State) -> None:
        systemctl_mqtt._dbus.service_manager.stop_unit(unit_name=self._unit_name)


class _MQTTActionRestartUnit(_MQTTAction):
    # pylint: disable=protected-access,too-few-public-methods
    def __init__(self, unit_name: str):
        self._unit_name = unit_name

    def trigger(self, state: _State) -> None:
        systemctl_mqtt._dbus.service_manager.restart_unit(unit_name=self._unit_name)


class _MQTTActionLockAllSessions(_MQTTAction):
    # pylint: disable=too-few-public-methods
    def trigger(self, state: _State) -> None:
        # pylint: disable=protected-access
        systemctl_mqtt._dbus.login_manager.lock_all_sessions()


class _MQTTActionSuspend(_MQTTAction):
    # pylint: disable=too-few-public-methods
    def trigger(self, state: _State) -> None:
        # pylint: disable=protected-access
        systemctl_mqtt._dbus.login_manager.suspend()


_MQTT_TOPIC_SUFFIX_ACTION_MAPPING = {
    "poweroff": _MQTTActionSchedulePoweroff(),
    "lock-all-sessions": _MQTTActionLockAllSessions(),
    "suspend": _MQTTActionSuspend(),
}


async def _mqtt_message_loop(*, state: _State, mqtt_client: aiomqtt.Client) -> None:
    action_by_topic: typing.Dict[str, _MQTTAction] = {}
    for topic_suffix, action in _MQTT_TOPIC_SUFFIX_ACTION_MAPPING.items():
        topic = state.mqtt_topic_prefix + "/" + topic_suffix
        _LOGGER.info("subscribing to %s", topic)
        await mqtt_client.subscribe(topic)
        action_by_topic[topic] = action

    for unit_name in state.controlled_system_unit_names:
        for topic_suffix, action_class in [
            ("start", _MQTTActionStartUnit),
            ("stop", _MQTTActionStopUnit),
            ("restart", _MQTTActionRestartUnit),
        ]:
            topic = (
                state.mqtt_topic_prefix
                + "/unit/system/"
                + unit_name
                + "/"
                + topic_suffix
            )
            _LOGGER.info("subscribing to %s", topic)
            await mqtt_client.subscribe(topic)
            action_by_topic[topic] = action_class(unit_name=unit_name)

    async for message in mqtt_client.messages:
        if message.retain:
            _LOGGER.info("ignoring retained message on topic %r", message.topic.value)
        else:
            _LOGGER.debug(
                "received message on topic %r: %r", message.topic.value, message.payload
            )
            action_by_topic[message.topic.value].trigger(state=state)


async def _dbus_signal_loop_preparing_for_shutdown(
    *,
    state: _State,
    mqtt_client: aiomqtt.Client,
    dbus_router: jeepney.io.asyncio.DBusRouter,
    bus_proxy: jeepney.io.asyncio.Proxy,
) -> None:
    preparing_for_shutdown_match_rule = (
        # pylint: disable=protected-access
        systemctl_mqtt._dbus.login_manager.get_login_manager_signal_match_rule(
            "PrepareForShutdown"
        )
    )
    assert await bus_proxy.AddMatch(preparing_for_shutdown_match_rule) == ()
    with dbus_router.filter(preparing_for_shutdown_match_rule) as queue:
        while True:
            message: jeepney.low_level.Message = await queue.get()
            (preparing_for_shutdown,) = message.body
            await state.preparing_for_shutdown_handler(
                active=preparing_for_shutdown, mqtt_client=mqtt_client
            )
            queue.task_done()


async def _get_unit_path(
    *, service_manager: jeepney.io.asyncio.Proxy, unit_name: str
) -> str:
    (path,) = await service_manager.GetUnit(name=unit_name)
    return path


async def _dbus_signal_loop_unit(  # pylint: disable=too-many-arguments
    *,
    state: _State,
    mqtt_client: aiomqtt.Client,
    dbus_router: jeepney.io.asyncio.DBusRouter,
    bus_proxy: jeepney.io.asyncio.Proxy,
    unit_name: str,
    unit_path: str,
) -> None:
    unit_proxy = jeepney.io.asyncio.Proxy(
        # pylint: disable=protected-access
        msggen=systemctl_mqtt._dbus.service_manager.Unit(object_path=unit_path),
        router=dbus_router,
    )
    unit_properties_changed_match_rule = jeepney.MatchRule(
        type="signal",
        interface="org.freedesktop.DBus.Properties",
        member="PropertiesChanged",
        path=unit_path,
    )
    assert (await bus_proxy.AddMatch(unit_properties_changed_match_rule)) == ()
    # > Table 1. Unit ACTIVE states …
    # > active	Started, bound, plugged in, …
    # > inactive	Stopped, unbound, unplugged, …
    # > failed	… process returned error code on exit, crashed, an operation
    # .         timed out, or after too many restarts).
    # > activating	Changing from inactive to active.
    # > deactivating	Changing from active to inactive.
    # > maintenance	Unit is inactive and … maintenance … in progress.
    # > reloading	Unit is active and it is reloading its configuration.
    # > refreshing	Unit is active and a new mount is being activated in its
    # .             namespace.
    # https://web.archive.org/web/20250101121304/https://www.freedesktop.org/software/systemd/man/latest/org.freedesktop.systemd1.html
    active_state_topic = state.get_system_unit_active_state_mqtt_topic(
        unit_name=unit_name
    )
    ((_, last_active_state),) = await unit_proxy.Get(property_name="ActiveState")
    await mqtt_client.publish(topic=active_state_topic, payload=last_active_state)
    with dbus_router.filter(unit_properties_changed_match_rule) as queue:
        while True:
            await queue.get()
            ((_, current_active_state),) = await unit_proxy.Get(
                property_name="ActiveState"
            )
            if current_active_state != last_active_state:
                await mqtt_client.publish(
                    topic=active_state_topic, payload=current_active_state
                )
                last_active_state = current_active_state
            queue.task_done()


async def _dbus_signal_loop(*, state: _State, mqtt_client: aiomqtt.Client) -> None:
    async with jeepney.io.asyncio.open_dbus_router(bus="SYSTEM") as router:
        # router: jeepney.io.asyncio.DBusRouter
        bus_proxy = jeepney.io.asyncio.Proxy(
            msggen=jeepney.bus_messages.message_bus, router=router
        )
        system_service_manager = jeepney.io.asyncio.Proxy(
            # pylint: disable=protected-access
            msggen=systemctl_mqtt._dbus.service_manager.ServiceManager(),
            router=router,
        )
        await asyncio.gather(
            *[
                _dbus_signal_loop_preparing_for_shutdown(
                    state=state,
                    mqtt_client=mqtt_client,
                    dbus_router=router,
                    bus_proxy=bus_proxy,
                )
            ]
            + [
                _dbus_signal_loop_unit(
                    state=state,
                    mqtt_client=mqtt_client,
                    dbus_router=router,
                    bus_proxy=bus_proxy,
                    unit_name=unit_name,
                    unit_path=await _get_unit_path(
                        service_manager=system_service_manager, unit_name=unit_name
                    ),
                )
                for unit_name in state.monitored_system_unit_names
            ],
            return_exceptions=False,
        )


async def _run(  # pylint: disable=too-many-arguments
    *,
    mqtt_host: str,
    mqtt_port: int,
    mqtt_username: typing.Optional[str],
    mqtt_password: typing.Optional[str],
    mqtt_topic_prefix: str,
    homeassistant_discovery_prefix: str,
    homeassistant_discovery_object_id: str,
    poweroff_delay: datetime.timedelta,
    monitored_system_unit_names: typing.List[str],
    controlled_system_unit_names: typing.List[str],
    mqtt_disable_tls: bool = False,
) -> None:
    state = _State(
        mqtt_topic_prefix=mqtt_topic_prefix,
        homeassistant_discovery_prefix=homeassistant_discovery_prefix,
        homeassistant_discovery_object_id=homeassistant_discovery_object_id,
        poweroff_delay=poweroff_delay,
        monitored_system_unit_names=monitored_system_unit_names,
        controlled_system_unit_names=controlled_system_unit_names,
    )
    _LOGGER.info(
        "connecting to MQTT broker %s:%d (TLS %s)",
        mqtt_host,
        mqtt_port,
        "disabled" if mqtt_disable_tls else "enabled",
    )
    if mqtt_password and not mqtt_username:
        raise ValueError("Missing MQTT username")
    async with aiomqtt.Client(  # raises aiomqtt.MqttError
        hostname=mqtt_host,
        port=mqtt_port,
        # > The settings [...] usually represent a higher security level than
        # > when calling the SSLContext constructor directly.
        # https://web.archive.org/web/20230714183106/https://docs.python.org/3/library/ssl.html
        tls_context=None if mqtt_disable_tls else ssl.create_default_context(),
        username=None if mqtt_username is None else mqtt_username,
        password=None if mqtt_password is None else mqtt_password,
        will=aiomqtt.Will(  # e.g. on SIGTERM & SIGKILL
            topic=state.mqtt_availability_topic,
            payload=_MQTT_PAYLOAD_NOT_AVAILABLE,
            retain=True,
        ),
    ) as mqtt_client:
        _LOGGER.debug("connected to MQTT broker %s:%d", mqtt_host, mqtt_port)
        if not state.shutdown_lock_acquired:
            state.acquire_shutdown_lock()
        await state.publish_homeassistant_device_config(mqtt_client=mqtt_client)
        await state.publish_preparing_for_shutdown(mqtt_client=mqtt_client)
        try:
            await mqtt_client.publish(
                topic=state.mqtt_availability_topic,
                payload=_MQTT_PAYLOAD_AVAILABLE,
                retain=True,
            )
            # asynpio.TaskGroup added in python3.11
            await asyncio.gather(
                _mqtt_message_loop(state=state, mqtt_client=mqtt_client),
                _dbus_signal_loop(state=state, mqtt_client=mqtt_client),
                return_exceptions=False,
            )
        finally:  # e.g. on SIGINT
            # https://web.archive.org/web/20250101080719/https://github.com/empicano/aiomqtt/issues/28
            await mqtt_client.publish(
                topic=state.mqtt_availability_topic,
                payload=_MQTT_PAYLOAD_NOT_AVAILABLE,
                retain=True,
            )


def _main() -> None:
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s:%(levelname)s:%(message)s",
        datefmt="%Y-%m-%dT%H:%M:%S%z",
    )
    argparser = argparse.ArgumentParser(
        description="MQTT client triggering & reporting shutdown on systemd-based systems",
    )
    argparser.add_argument(
        "--log-level",
        choices=_ARGUMENT_LOG_LEVEL_MAPPING.keys(),
        default="info",
        help="log level (default: %(default)s)",
    )
    argparser.add_argument("--mqtt-host", type=str, required=True)
    argparser.add_argument(
        "--mqtt-port",
        type=int,
        help=f"default {_MQTT_DEFAULT_TLS_PORT} ({_MQTT_DEFAULT_PORT} with --mqtt-disable-tls)",
    )
    argparser.add_argument("--mqtt-username", type=str)
    argparser.add_argument("--mqtt-disable-tls", action="store_true")
    password_argument_group = argparser.add_mutually_exclusive_group()
    password_argument_group.add_argument("--mqtt-password", type=str)
    password_argument_group.add_argument(
        "--mqtt-password-file",
        type=pathlib.Path,
        metavar="PATH",
        dest="mqtt_password_path",
        help="stripping trailing newline",
    )
    argparser.add_argument(
        "--mqtt-topic-prefix",
        type=str,
        # pylint: disable=protected-access
        default="systemctl/" + systemctl_mqtt._utils.get_hostname(),
        help="default: %(default)s",
    )
    # https://www.home-assistant.io/docs/mqtt/discovery/#discovery_prefix
    argparser.add_argument(
        "--homeassistant-discovery-prefix",
        type=str,
        default="homeassistant",
        help="home assistant's prefix for discovery topics" + " (default: %(default)s)",
    )
    argparser.add_argument(
        "--homeassistant-discovery-object-id",
        type=str,
        # pylint: disable=protected-access
        default=systemctl_mqtt._homeassistant.get_default_discovery_object_id(),
        help="part of discovery topic (default: %(default)s)",
    )
    argparser.add_argument(
        "--poweroff-delay-seconds", type=float, default=4.0, help="default: %(default)s"
    )
    argparser.add_argument(
        "--monitor-system-unit",
        type=str,
        metavar="UNIT_NAME",
        dest="monitored_system_unit_names",
        action="append",
        help="e.g. --monitor-system-unit ssh.service --monitor-system-unit custom.service",
    )
    argparser.add_argument(
        "--control-system-unit",
        type=str,
        metavar="UNIT_NAME",
        dest="controlled_system_unit_names",
        action="append",
        help="e.g. --control-system-unit ansible-pull.service --control-system-unit custom.service",
    )
    args = argparser.parse_args()
    logging.root.setLevel(_ARGUMENT_LOG_LEVEL_MAPPING[args.log_level])
    if args.mqtt_port:
        mqtt_port = args.mqtt_port
    elif args.mqtt_disable_tls:
        mqtt_port = _MQTT_DEFAULT_PORT
    else:
        mqtt_port = _MQTT_DEFAULT_TLS_PORT
    if args.mqtt_password_path:
        # .read_text() replaces \r\n with \n
        mqtt_password = args.mqtt_password_path.read_bytes().decode()
        if mqtt_password.endswith("\r\n"):
            mqtt_password = mqtt_password[:-2]
        elif mqtt_password.endswith("\n"):
            mqtt_password = mqtt_password[:-1]
    else:
        mqtt_password = args.mqtt_password
    # pylint: disable=protected-access
    if not systemctl_mqtt._homeassistant.validate_discovery_object_id(
        args.homeassistant_discovery_object_id
    ):
        raise ValueError(
            # pylint: disable=protected-access
            "invalid home assistant discovery object id"
            f" {args.homeassistant_discovery_object_id!r} (length >= 1"
            ", allowed characters:"
            f" {systemctl_mqtt._homeassistant.NODE_ID_ALLOWED_CHARS})"
            "\nchange --homeassistant-discovery-object-id"
        )
    asyncio.run(
        _run(
            mqtt_host=args.mqtt_host,
            mqtt_port=mqtt_port,
            mqtt_disable_tls=args.mqtt_disable_tls,
            mqtt_username=args.mqtt_username,
            mqtt_password=mqtt_password,
            mqtt_topic_prefix=args.mqtt_topic_prefix,
            homeassistant_discovery_prefix=args.homeassistant_discovery_prefix,
            homeassistant_discovery_object_id=args.homeassistant_discovery_object_id,
            poweroff_delay=datetime.timedelta(seconds=args.poweroff_delay_seconds),
            monitored_system_unit_names=args.monitored_system_unit_names or [],
            controlled_system_unit_names=args.controlled_system_unit_names or [],
        )
    )



# systemctl-mqtt - MQTT client triggering & reporting shutdown on systemd-based systems
#
# Copyright (C) 2020 Fabian Peter Hammerle <fabian@hammerle.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime
import logging
import ssl
import unittest.mock

import aiomqtt
import jeepney.fds
import jeepney.low_level
import pytest

import systemctl_mqtt

# pylint: disable=protected-access,too-many-positional-arguments


@pytest.mark.asyncio
@pytest.mark.parametrize("mqtt_host", ["mqtt-broker.local"])
@pytest.mark.parametrize("mqtt_port", [1883])
@pytest.mark.parametrize("mqtt_topic_prefix", ["systemctl/host", "system/command"])
@pytest.mark.parametrize("homeassistant_discovery_prefix", ["homeassistant"])
@pytest.mark.parametrize("homeassistant_discovery_object_id", ["host", "node"])
async def test__run(
    caplog,
    mqtt_host,
    mqtt_port,
    mqtt_topic_prefix,
    homeassistant_discovery_prefix,
    homeassistant_discovery_object_id,
):
    # pylint: disable=too-many-locals,too-many-arguments
    caplog.set_level(logging.DEBUG)
    login_manager_mock = unittest.mock.MagicMock()
    with unittest.mock.patch(
        "aiomqtt.Client", autospec=False
    ) as mqtt_client_class_mock, unittest.mock.patch(
        "systemctl_mqtt._dbus.login_manager.get_login_manager_proxy",
        return_value=login_manager_mock,
    ), unittest.mock.patch(
        "systemctl_mqtt._dbus_signal_loop"
    ) as dbus_signal_loop_mock:
        login_manager_mock.Inhibit.return_value = (jeepney.fds.FileDescriptor(-1),)
        login_manager_mock.Get.return_value = (("b", False),)
        await systemctl_mqtt._run(
            mqtt_host=mqtt_host,
            mqtt_port=mqtt_port,
            mqtt_username=None,
            mqtt_password=None,
            mqtt_topic_prefix=mqtt_topic_prefix,
            homeassistant_discovery_prefix=homeassistant_discovery_prefix,
            homeassistant_discovery_object_id=homeassistant_discovery_object_id,
            poweroff_delay=datetime.timedelta(),
            monitored_system_unit_names=[],
            controlled_system_unit_names=[],
        )
    assert caplog.records[0].levelno == logging.INFO
    assert caplog.records[0].message == (
        f"connecting to MQTT broker {mqtt_host}:{mqtt_port} (TLS enabled)"
    )
    mqtt_client_class_mock.assert_called_once()
    _, mqtt_client_init_kwargs = mqtt_client_class_mock.call_args
    assert mqtt_client_init_kwargs.pop("hostname") == mqtt_host
    assert mqtt_client_init_kwargs.pop("port") == mqtt_port
    assert isinstance(mqtt_client_init_kwargs.pop("tls_context"), ssl.SSLContext)
    assert mqtt_client_init_kwargs.pop("username") is None
    assert mqtt_client_init_kwargs.pop("password") is None
    assert mqtt_client_init_kwargs.pop("will") == aiomqtt.Will(
        topic=mqtt_topic_prefix + "/status",
        payload="offline",
        qos=0,
        retain=True,
        properties=None,
    )
    assert not mqtt_client_init_kwargs
    login_manager_mock.Inhibit.assert_called_once_with(
        what="shutdown",
        who="systemctl-mqtt",
        why="Report shutdown via MQTT",
        mode="delay",
    )
    login_manager_mock.Get.assert_called_once_with("PreparingForShutdown")
    async with mqtt_client_class_mock() as mqtt_client_mock:
        pass
    assert mqtt_client_mock.publish.call_count == 4
    assert (
        mqtt_client_mock.publish.call_args_list[0][1]["topic"]
        == f"{homeassistant_discovery_prefix}/device/{homeassistant_discovery_object_id}/config"
    )
    assert mqtt_client_mock.publish.call_args_list[1] == unittest.mock.call(
        topic=mqtt_topic_prefix + "/preparing-for-shutdown",
        payload="false",
        retain=False,
    )
    assert mqtt_client_mock.publish.call_args_list[2][1] == {
        "topic": mqtt_topic_prefix + "/status",
        "payload": "online",
        "retain": True,
    }
    assert mqtt_client_mock.publish.call_args_list[3][1] == {
        "topic": mqtt_topic_prefix + "/status",
        "payload": "offline",
        "retain": True,
    }
    assert sorted(mqtt_client_mock.subscribe.call_args_list) == [
        unittest.mock.call(mqtt_topic_prefix + "/lock-all-sessions"),
        unittest.mock.call(mqtt_topic_prefix + "/poweroff"),
        unittest.mock.call(mqtt_topic_prefix + "/suspend"),
    ]
    assert caplog.records[1].levelno == logging.DEBUG
    assert (
        caplog.records[1].message == f"connected to MQTT broker {mqtt_host}:{mqtt_port}"
    )
    assert caplog.records[2].levelno == logging.DEBUG
    assert caplog.records[2].message == "acquired shutdown inhibitor lock"
    assert caplog.records[3].levelno == logging.DEBUG
    assert (
        caplog.records[3].message
        == "publishing home assistant config on "
        + homeassistant_discovery_prefix
        + "/device/"
        + homeassistant_discovery_object_id
        + "/config"
    )
    assert caplog.records[4].levelno == logging.INFO
    assert (
        caplog.records[4].message
        == f"publishing 'false' on {mqtt_topic_prefix}/preparing-for-shutdown"
    )
    assert all(r.levelno == logging.INFO for r in caplog.records[5::2])
    assert {r.message for r in caplog.records[5:]} == {
        f"subscribing to {mqtt_topic_prefix}/{s}"
        for s in ("poweroff", "lock-all-sessions", "suspend")
    }
    dbus_signal_loop_mock.assert_awaited_once()


@pytest.mark.asyncio
@pytest.mark.parametrize("mqtt_host", ["mqtt-broker.local"])
@pytest.mark.parametrize("mqtt_port", [1833])
@pytest.mark.parametrize("mqtt_disable_tls", [True, False])
async def test__run_tls(caplog, mqtt_host, mqtt_port, mqtt_disable_tls):
    caplog.set_level(logging.INFO)
    with unittest.mock.patch(
        "aiomqtt.Client"
    ) as mqtt_client_class_mock, unittest.mock.patch(
        "systemctl_mqtt._dbus_signal_loop"
    ) as dbus_signal_loop_mock:
        await systemctl_mqtt._run(
            mqtt_host=mqtt_host,
            mqtt_port=mqtt_port,
            mqtt_disable_tls=mqtt_disable_tls,
            mqtt_username=None,
            mqtt_password=None,
            mqtt_topic_prefix="systemctl/hosts",
            homeassistant_discovery_prefix="homeassistant",
            homeassistant_discovery_object_id="host",
            poweroff_delay=datetime.timedelta(),
            monitored_system_unit_names=[],
            controlled_system_unit_names=[],
        )
    mqtt_client_class_mock.assert_called_once()
    _, mqtt_client_init_kwargs = mqtt_client_class_mock.call_args
    assert mqtt_client_init_kwargs.pop("hostname") == mqtt_host
    assert mqtt_client_init_kwargs.pop("port") == mqtt_port
    if mqtt_disable_tls:
        assert mqtt_client_init_kwargs.pop("tls_context") is None
    else:
        assert isinstance(mqtt_client_init_kwargs.pop("tls_context"), ssl.SSLContext)
    assert set(mqtt_client_init_kwargs.keys()) == {"username", "password", "will"}
    assert caplog.records[0].levelno == logging.INFO
    assert caplog.records[0].message == (
        f"connecting to MQTT broker {mqtt_host}:{mqtt_port}"
        f" (TLS {'disabled' if mqtt_disable_tls else 'enabled'})"
    )
    dbus_signal_loop_mock.assert_awaited_once()


@pytest.mark.asyncio
async def test__run_tls_default():
    with unittest.mock.patch(
        "aiomqtt.Client"
    ) as mqtt_client_class_mock, unittest.mock.patch(
        "systemctl_mqtt._dbus_signal_loop"
    ) as dbus_signal_loop_mock:
        await systemctl_mqtt._run(
            mqtt_host="mqtt-broker.local",
            mqtt_port=1883,
            # mqtt_disable_tls default,
            mqtt_username=None,
            mqtt_password=None,
            mqtt_topic_prefix="systemctl/hosts",
            homeassistant_discovery_prefix="homeassistant",
            homeassistant_discovery_object_id="host",
            poweroff_delay=datetime.timedelta(),
            monitored_system_unit_names=[],
            controlled_system_unit_names=[],
        )
    mqtt_client_class_mock.assert_called_once()
    # enabled by default
    assert isinstance(
        mqtt_client_class_mock.call_args[1]["tls_context"], ssl.SSLContext
    )
    dbus_signal_loop_mock.assert_awaited_once()


@pytest.mark.asyncio
@pytest.mark.parametrize("mqtt_host", ["mqtt-broker.local"])
@pytest.mark.parametrize("mqtt_port", [1883])
@pytest.mark.parametrize("mqtt_username", ["me"])
@pytest.mark.parametrize("mqtt_password", [None, "secret"])
@pytest.mark.parametrize("mqtt_topic_prefix", ["systemctl/host"])
async def test__run_authentication(
    mqtt_host, mqtt_port, mqtt_username, mqtt_password, mqtt_topic_prefix
):
    with unittest.mock.patch(
        "aiomqtt.Client"
    ) as mqtt_client_class_mock, unittest.mock.patch(
        "systemctl_mqtt._dbus_signal_loop"
    ) as dbus_signal_loop_mock:
        await systemctl_mqtt._run(
            mqtt_host=mqtt_host,
            mqtt_port=mqtt_port,
            mqtt_username=mqtt_username,
            mqtt_password=mqtt_password,
            mqtt_topic_prefix=mqtt_topic_prefix,
            homeassistant_discovery_prefix="discovery-prefix",
            homeassistant_discovery_object_id="node-id",
            poweroff_delay=datetime.timedelta(),
            monitored_system_unit_names=[],
            controlled_system_unit_names=[],
        )
    mqtt_client_class_mock.assert_called_once()
    _, mqtt_client_init_kwargs = mqtt_client_class_mock.call_args
    assert mqtt_client_init_kwargs["username"] == mqtt_username
    if mqtt_password:
        assert mqtt_client_init_kwargs["password"] == mqtt_password
    else:
        assert mqtt_client_init_kwargs["password"] is None
    dbus_signal_loop_mock.assert_awaited_once()


@pytest.mark.asyncio
@pytest.mark.parametrize("mqtt_host", ["mqtt-broker.local"])
@pytest.mark.parametrize("mqtt_port", [1883])
@pytest.mark.parametrize("mqtt_password", ["secret"])
async def test__run_authentication_missing_username(
    mqtt_host: str, mqtt_port: int, mqtt_password: str
) -> None:
    with unittest.mock.patch("aiomqtt.Client"), unittest.mock.patch(
        "systemctl_mqtt._dbus.login_manager.get_login_manager_proxy"
    ), unittest.mock.patch("systemctl_mqtt._dbus_signal_loop") as dbus_signal_loop_mock:
        with pytest.raises(ValueError, match=r"^Missing MQTT username$"):
            await systemctl_mqtt._run(
                mqtt_host=mqtt_host,
                mqtt_port=mqtt_port,
                mqtt_username=None,
                mqtt_password=mqtt_password,
                mqtt_topic_prefix="prefix",
                homeassistant_discovery_prefix="discovery-prefix",
                homeassistant_discovery_object_id="node-id",
                poweroff_delay=datetime.timedelta(),
                monitored_system_unit_names=[],
                controlled_system_unit_names=[],
            )
    dbus_signal_loop_mock.assert_not_called()


@pytest.mark.asyncio
@pytest.mark.parametrize("mqtt_topic_prefix", ["systemctl/host"])
async def test__run_sigint(mqtt_topic_prefix: str):
    login_manager_mock = unittest.mock.MagicMock()
    with unittest.mock.patch(
        "aiomqtt.Client", autospec=False
    ) as mqtt_client_class_mock, unittest.mock.patch(
        "systemctl_mqtt._dbus.login_manager.get_login_manager_proxy",
        return_value=login_manager_mock,
    ), unittest.mock.patch(
        "asyncio.gather", side_effect=KeyboardInterrupt
    ):
        login_manager_mock.Inhibit.return_value = (jeepney.fds.FileDescriptor(-1),)
        login_manager_mock.Get.return_value = (("b", False),)
        with pytest.raises(KeyboardInterrupt):
            await systemctl_mqtt._run(
                mqtt_host="mqtt-broker.local",
                mqtt_port=1883,
                mqtt_username=None,
                mqtt_password=None,
                mqtt_topic_prefix=mqtt_topic_prefix,
                homeassistant_discovery_prefix="homeassistant",
                homeassistant_discovery_object_id="host",
                poweroff_delay=datetime.timedelta(),
                monitored_system_unit_names=[],
                controlled_system_unit_names=[],
            )
    async with mqtt_client_class_mock() as mqtt_client_mock:
        pass
    assert mqtt_client_mock.publish.call_count == 4
    assert mqtt_client_mock.publish.call_args_list[0][1]["topic"].endswith("/config")
    assert mqtt_client_mock.publish.call_args_list[1][1]["topic"].endswith(
        "/preparing-for-shutdown"
    )
    assert mqtt_client_mock.publish.call_args_list[2][1] == {
        "topic": mqtt_topic_prefix + "/status",
        "payload": "online",
        "retain": True,
    }
    assert mqtt_client_mock.publish.call_args_list[3][1] == {
        "topic": mqtt_topic_prefix + "/status",
        "payload": "offline",
        "retain": True,
    }


@pytest.mark.asyncio
@pytest.mark.filterwarnings("ignore:coroutine '_dbus_signal_loop' was never awaited")
@pytest.mark.filterwarnings("ignore:coroutine '_mqtt_message_loop' was never awaited")
@pytest.mark.parametrize("mqtt_topic_prefix", ["systemctl/host", "system/command"])
async def test__mqtt_message_loop_trigger_poweroff(
    caplog: pytest.LogCaptureFixture, mqtt_topic_prefix: str
) -> None:
    state = systemctl_mqtt._State(
        mqtt_topic_prefix=mqtt_topic_prefix,
        homeassistant_discovery_prefix="homeassistant",
        homeassistant_discovery_object_id="whatever",
        poweroff_delay=datetime.timedelta(seconds=21),
        monitored_system_unit_names=[],
        controlled_system_unit_names=[],
    )
    mqtt_client_mock = unittest.mock.AsyncMock()
    mqtt_client_mock.messages.__aiter__.return_value = [
        aiomqtt.Message(
            topic=mqtt_topic_prefix + "/poweroff",
            payload=b"some-payload",
            qos=0,
            retain=False,
            mid=42 // 2,
            properties=None,
        )
    ]
    with unittest.mock.patch(
        "systemctl_mqtt._dbus.login_manager.schedule_shutdown"
    ) as schedule_shutdown_mock, caplog.at_level(logging.DEBUG):
        await systemctl_mqtt._mqtt_message_loop(
            state=state, mqtt_client=mqtt_client_mock
        )
    assert sorted(mqtt_client_mock.subscribe.await_args_list) == [
        unittest.mock.call(mqtt_topic_prefix + "/lock-all-sessions"),
        unittest.mock.call(mqtt_topic_prefix + "/poweroff"),
        unittest.mock.call(mqtt_topic_prefix + "/suspend"),
    ]
    schedule_shutdown_mock.assert_called_once_with(
        action="poweroff", delay=datetime.timedelta(seconds=21)
    )
    assert [
        t for t in caplog.record_tuples[2:] if not t[2].startswith("subscribing to ")
    ] == [
        (
            "systemctl_mqtt",
            logging.DEBUG,
            f"received message on topic '{mqtt_topic_prefix}/poweroff': b'some-payload'",
        ),
    ]


@pytest.mark.asyncio
@pytest.mark.parametrize("mqtt_topic_prefix", ["systemctl/host"])
async def test__mqtt_message_loop_retained(
    caplog: pytest.LogCaptureFixture, mqtt_topic_prefix: str
) -> None:
    state = systemctl_mqtt._State(
        mqtt_topic_prefix=mqtt_topic_prefix,
        homeassistant_discovery_prefix="homeassistant",
        homeassistant_discovery_object_id="whatever",
        poweroff_delay=datetime.timedelta(seconds=21),
        monitored_system_unit_names=[],
        controlled_system_unit_names=[],
    )
    mqtt_client_mock = unittest.mock.AsyncMock()
    mqtt_client_mock.messages.__aiter__.return_value = [
        aiomqtt.Message(
            topic=mqtt_topic_prefix + "/poweroff",
            payload=b"some-payload",
            qos=0,
            retain=True,
            mid=42 // 2,
            properties=None,
        )
    ]
    with unittest.mock.patch(
        "systemctl_mqtt._dbus.login_manager.schedule_shutdown"
    ) as schedule_shutdown_mock, caplog.at_level(logging.DEBUG):
        await systemctl_mqtt._mqtt_message_loop(
            state=state, mqtt_client=mqtt_client_mock
        )
    schedule_shutdown_mock.assert_not_called()
    assert [
        t for t in caplog.record_tuples[2:] if not t[2].startswith("subscribing to ")
    ] == [
        (
            "systemctl_mqtt",
            logging.INFO,
            "ignoring retained message on topic 'systemctl/host/poweroff'",
        ),
    ]


@pytest.mark.parametrize("mqtt_topic_prefix", ["systemctl/host", "systemd/raspberrypi"])
@pytest.mark.parametrize("unit_name", ["foo.service", "bar.service"])
def test_state_get_system_unit_active_state_mqtt_topic(
    mqtt_topic_prefix: str, unit_name: str
) -> None:
    state = systemctl_mqtt._State(
        mqtt_topic_prefix=mqtt_topic_prefix,
        homeassistant_discovery_prefix="homeassistant",
        homeassistant_discovery_object_id="whatever",
        poweroff_delay=datetime.timedelta(seconds=21),
        monitored_system_unit_names=[],
        controlled_system_unit_names=[],
    )
    assert (
        state.get_system_unit_active_state_mqtt_topic(unit_name=unit_name)
        == f"{mqtt_topic_prefix}/unit/system/{unit_name}/active-state"
    )


@pytest.mark.asyncio
@pytest.mark.filterwarnings("ignore:coroutine '_dbus_signal_loop' was never awaited")
@pytest.mark.filterwarnings("ignore:coroutine '_mqtt_message_loop' was never awaited")
@pytest.mark.parametrize("mqtt_topic_prefix", ["systemctl/host"])
@pytest.mark.parametrize("unit_name", ["foo.service", "bar.service"])
@pytest.mark.parametrize("action", ["restart", "start", "stop"])
async def test__mqtt_message_loop_triggers_unit_action(
    caplog: pytest.LogCaptureFixture,
    mqtt_topic_prefix: str,
    unit_name: str,
    action: str,
) -> None:
    state = systemctl_mqtt._State(
        mqtt_topic_prefix=mqtt_topic_prefix,
        homeassistant_discovery_prefix="homeassistant",
        homeassistant_discovery_object_id="whatever",
        poweroff_delay=datetime.timedelta(seconds=21),
        monitored_system_unit_names=[],
        controlled_system_unit_names=[unit_name],
    )

    mqtt_client_mock = unittest.mock.AsyncMock()
    topic = f"{mqtt_topic_prefix}/unit/system/{unit_name}/{action}"
    mqtt_client_mock.messages.__aiter__.return_value = [
        aiomqtt.Message(
            topic=topic,
            payload=b"some-payload",
            qos=0,
            retain=False,
            mid=42 // 2,
            properties=None,
        )
    ]

    with unittest.mock.patch(
        f"systemctl_mqtt._dbus.service_manager.{action}_unit"
    ) as trigger_service_mock, caplog.at_level(logging.DEBUG):
        await systemctl_mqtt._mqtt_message_loop(
            state=state, mqtt_client=mqtt_client_mock
        )

    # check subscription
    assert unittest.mock.call(topic) in mqtt_client_mock.subscribe.await_args_list

    # check correct action method called
    trigger_service_mock.assert_called_once_with(unit_name=unit_name)

    # check logs (skip "subscribing to ..." chatter)
    assert [
        t for t in caplog.record_tuples[2:] if not t[2].startswith("subscribing to ")
    ] == [
        (
            "systemctl_mqtt",
            logging.DEBUG,
            f"received message on topic '{topic}': b'some-payload'",
        ),
    ]



# systemctl-mqtt - MQTT client triggering & reporting shutdown on systemd-based systems
#
# Copyright (C) 2024 Fabian Peter Hammerle <fabian@hammerle.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import contextlib
import datetime
import typing
import unittest.mock

import pytest
from jeepney.low_level import HeaderFields, Message

import systemctl_mqtt._dbus.login_manager

# pylint: disable=protected-access


@contextlib.contextmanager
def mock_open_dbus_connection() -> typing.Iterator[unittest.mock.MagicMock]:
    with unittest.mock.patch("jeepney.io.blocking.open_dbus_connection") as mock:
        yield mock.return_value


@pytest.mark.parametrize(
    ("member", "signature", "kwargs", "body"),
    [
        ("ListInhibitors", None, {}, ()),
        ("LockSessions", None, {}, ()),
        ("CanPowerOff", None, {}, ()),
        (
            "ScheduleShutdown",
            "st",
            {
                "action": "poweroff",
                "time": datetime.datetime(
                    1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc
                ),
            },
            ("poweroff", 0),
        ),
        ("Suspend", "b", {"interactive": True}, (True,)),
        (
            "Inhibit",
            "ssss",
            {"what": "poweroff", "who": "me", "why": "fixing bugs", "mode": "block"},
            ("poweroff", "me", "fixing bugs", "block"),
        ),
    ],
)
def test_method(
    member: str,
    signature: typing.Optional[str],
    kwargs: typing.Dict[str, typing.Any],
    body: typing.Tuple[typing.Any],
) -> None:
    with mock_open_dbus_connection() as dbus_connection_mock:
        proxy = systemctl_mqtt._dbus.login_manager.get_login_manager_proxy()
    getattr(proxy, member)(**kwargs)
    dbus_connection_mock.send_and_get_reply.assert_called_once()
    message: Message = dbus_connection_mock.send_and_get_reply.call_args[0][0]
    if signature:
        assert message.header.fields.pop(HeaderFields.signature) == signature
    assert message.header.fields == {
        HeaderFields.path: "/org/freedesktop/login1",
        HeaderFields.destination: "org.freedesktop.login1",
        HeaderFields.interface: "org.freedesktop.login1.Manager",
        HeaderFields.member: member,
    }
    assert message.body == body


@pytest.mark.parametrize("property_name", ["HandlePowerKey", "Docked"])
def test_get(property_name: str) -> None:
    with mock_open_dbus_connection() as dbus_connection_mock:
        proxy = systemctl_mqtt._dbus.login_manager.get_login_manager_proxy()
    proxy.Get(property_name=property_name)
    dbus_connection_mock.send_and_get_reply.assert_called_once()
    message: Message = dbus_connection_mock.send_and_get_reply.call_args[0][0]
    assert message.header.fields == {
        HeaderFields.path: "/org/freedesktop/login1",
        HeaderFields.destination: "org.freedesktop.login1",
        HeaderFields.interface: "org.freedesktop.DBus.Properties",
        HeaderFields.member: "Get",
        HeaderFields.signature: "ss",
    }
    assert message.body == ("org.freedesktop.login1.Manager", property_name)



# systemctl-mqtt - MQTT client triggering & reporting shutdown on systemd-based systems
#
# Copyright (C) 2020 Fabian Peter Hammerle <fabian@hammerle.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import pathlib

import setuptools

_REPO_URL = "https://github.com/fphammerle/systemctl-mqtt"

setuptools.setup(
    name="systemctl-mqtt",
    packages=setuptools.find_packages(),
    description="MQTT client triggering & reporting shutdown on systemd-based systems",
    long_description=pathlib.Path(__file__).parent.joinpath("README.md").read_text(),
    long_description_content_type="text/markdown",
    author="Fabian Peter Hammerle",
    author_email="fabian@hammerle.me",
    url=_REPO_URL,
    project_urls={"Changelog": _REPO_URL + "/blob/master/CHANGELOG.md"},
    license="GPLv3+",
    keywords=[
        "IoT",
        "automation",
        "home-assistant",
        "home-automation",
        "lock",
        "mqtt",
        "shutdown",
        "systemd",
    ],
    classifiers=[
        # https://pypi.org/classifiers/
        "Development Status :: 4 - Beta",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: POSIX :: Linux",
        # .github/workflows/python.yml
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Home Automation",
    ],
    entry_points={
        "console_scripts": [
            "systemctl-mqtt = systemctl_mqtt:_main",
        ]
    },
    # >=3.6 variable type hints, f-strings & * to force keyword-only arguments
    # >=3.8 importlib.metadata
    python_requires=">=3.9",  # <3.9 untested
    # > Currently, the only main loop supported by dbus-python is GLib.
    # https://web.archive.org/web/20241228081405/https://dbus.freedesktop.org/doc/dbus-python/tutorial.html#setting-up-an-event-loop
    # PyGObject depends on pycairo
    # > When pip-installing systemctl-mqtt on a system without graphics it
    # > fails as pycairo fails building.
    # https://web.archive.org/web/20241228083145/https://github.com/fphammerle/systemctl-mqtt/issues/39
    # > Jeepney is a pure Python D-Bus module. It consists of an IO-free core
    # > implementing the protocol, and integrations for both blocking I/O and
    # > for different asynchronous frameworks.
    # https://web.archive.org/web/20241206000411/https://www.freedesktop.org/wiki/Software/DBusBindings/
    install_requires=["aiomqtt>=2,<3", "jeepney>=0.8,<1.0"],
    setup_requires=["setuptools_scm"],
    tests_require=["pytest"],
)
